-----------------------------------------------------------------------------------------------
-- Client Lua Script for LocalQuestTrackerAdvanced
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------

require "Window"
require "QuestLib"


local LocalQuestTrackerAdvanced = {}
local VILib = _G["GeminiPackages"]:GetPackage("VILib-1.2")
local version = "v1.1.3"
local GeminiLogging = nil
local log = nil

-----------------------------------------------------------------------------------------------
-- New Variables
-----------------------------------------------------------------------------------------------

local SettingsWindowRef, SettingsRef, QuestTrackerRef

local tAnchors = {
	{sLabel = "Center", tPoints = {0.5, 0.5, 0.5, 0.5}, tOffsetMults = {-0.5, -0.5, 0.5, 0.5}},
	{sLabel = "Left", tPoints = {0, 0.5, 0, 0.5}, tOffsetMults = {0, -0.5, 1, 0.5}},
	{sLabel = "Top Left", tPoints = {0, 0, 0, 0}, tOffsetMults = {0, 0, 1, 1}},
	{sLabel = "Top", tPoints = {0.5, 0, 0.5, 0}, tOffsetMults = {-0.5, 0, 0.5, 1}},
	{sLabel = "Top Right", tPoints = {1, 0, 1, 0}, tOffsetMults = {-1, 0, 0, 1}},
	{sLabel = "Right", tPoints = {1, 0.5, 1, 0.5}, tOffsetMults = {-1, -0.5, 0, 0.5}},
	{sLabel = "Bottom Right", tPoints = {1, 1, 1, 1}, tOffsetMults = {-1, -1, 0, 0}},
	{sLabel = "Bottom", tPoints = {0.5, 1, 0.5, 1}, tOffsetMults = {-0.5, -1, 0.5, 0}},
	{sLabel = "Bottom Left", tPoints = {0, 1, 0, 1}, tOffsetMults = {0, -1, 1, 0}},
}

local tDefaultSettings = {
	bFreeScaling				= false,
	nWidth						= 300,
	nHeight						= 600,
	bFreePositioning			= false,
	bTopAlign					= false,
	tWindowPositions = {
		tMainOffsets = {
			{0, 0},
			{0, 0},
			{0, 0},
			{0, 0},
			{0, 0},
			{0, 0},
			{30, -90},
			{0, 0},
			{0, 0},
		},
		tSettings = {},
		nMainAnchor = 7,
	},

	bAutoCollapseEpisodes		= true,
	bAutoCollapseQuests			= true,
	nTimeout					= 60,
	bAutoHideEpisodes			= false,
	bAutoTrackWorldStory		= true,
	bAutoTrackZoneStory		= true,
	bAutoTrackRegionalStory		= true,
	bAutoTrackTasks				= false,
	bAutoShowHideTradeskills	= true,
	bAutoShowHideEpisodes		= false,

	bFirstRun					= true,
	bHideWhileChallenge			= false,
}

-----------------------------------------------------------------------------------------------
-- Standard Variables
-----------------------------------------------------------------------------------------------

local knMaxZombieEventCount 	= 7
local knQuestProgBarFadeoutTime = 10
local kstrPublicEventMarker 	= "Public Event"
local ktNumbersToLetters		=
{
	Apollo.GetString("QuestTracker_ObjectiveA"),
	Apollo.GetString("QuestTracker_ObjectiveB"),
	Apollo.GetString("QuestTracker_ObjectiveC"),
	Apollo.GetString("QuestTracker_ObjectiveD"),
	Apollo.GetString("QuestTracker_ObjectiveE"),
	Apollo.GetString("QuestTracker_ObjectiveF"),
	Apollo.GetString("QuestTracker_ObjectiveG"),
	Apollo.GetString("QuestTracker_ObjectiveH"),
	Apollo.GetString("QuestTracker_ObjectiveI"),
	Apollo.GetString("QuestTracker_ObjectiveJ"),
	Apollo.GetString("QuestTracker_ObjectiveK"),
	Apollo.GetString("QuestTracker_ObjectiveL")
}
local karPathToString =
{
	[PlayerPathLib.PlayerPathType_Soldier] 		= Apollo.GetString("CRB_Soldier"),
	[PlayerPathLib.PlayerPathType_Settler] 		= Apollo.GetString("CRB_Settler"),
	[PlayerPathLib.PlayerPathType_Scientist] 	= Apollo.GetString("CRB_Scientist"),
	[PlayerPathLib.PlayerPathType_Explorer] 	= Apollo.GetString("CRB_Explorer")
}

local kstrRed 		= "ffff4c4c"
local kstrGreen 	= "ff2fdc02"
local kstrYellow 	= "fffffc00"
local kstrLightGrey = "ffb4b4b4"
local kstrHighlight = "ffffe153"

local ktConToColor =
{
	[0] 												= "ffffffff",
	[Unit.CodeEnumLevelDifferentialAttribute.Grey] 		= "ff9aaea3",
	[Unit.CodeEnumLevelDifferentialAttribute.Green] 	= "ff37ff00",
	[Unit.CodeEnumLevelDifferentialAttribute.Cyan] 		= "ff46ffff",
	[Unit.CodeEnumLevelDifferentialAttribute.Blue] 		= "ff3052fc",
	[Unit.CodeEnumLevelDifferentialAttribute.White] 	= "ffffffff",
	[Unit.CodeEnumLevelDifferentialAttribute.Yellow] 	= "ffffd400", -- Yellow
	[Unit.CodeEnumLevelDifferentialAttribute.Orange] 	= "ffff6a00", -- Orange
	[Unit.CodeEnumLevelDifferentialAttribute.Red] 		= "ffff0000", -- Red
	[Unit.CodeEnumLevelDifferentialAttribute.Magenta] 	= "fffb00ff", -- Purp
}

local ktConToString =
{
	[0] 												= Apollo.GetString("Unknown_Unit"),
	[Unit.CodeEnumLevelDifferentialAttribute.Grey] 		= Apollo.GetString("QuestLog_Trivial"),
	[Unit.CodeEnumLevelDifferentialAttribute.Green] 	= Apollo.GetString("QuestLog_Easy"),
	[Unit.CodeEnumLevelDifferentialAttribute.Cyan] 		= Apollo.GetString("QuestLog_Simple"),
	[Unit.CodeEnumLevelDifferentialAttribute.Blue] 		= Apollo.GetString("QuestLog_Standard"),
	[Unit.CodeEnumLevelDifferentialAttribute.White] 	= Apollo.GetString("QuestLog_Average"),
	[Unit.CodeEnumLevelDifferentialAttribute.Yellow] 	= Apollo.GetString("QuestLog_Moderate"),
	[Unit.CodeEnumLevelDifferentialAttribute.Orange] 	= Apollo.GetString("QuestLog_Tough"),
	[Unit.CodeEnumLevelDifferentialAttribute.Red] 		= Apollo.GetString("QuestLog_Hard"),
	[Unit.CodeEnumLevelDifferentialAttribute.Magenta] 	= Apollo.GetString("QuestLog_Impossible")
}

local ktPvPEventTypes =
{
	[PublicEvent.PublicEventType_PVP_Arena] 					= true,
	[PublicEvent.PublicEventType_PVP_Warplot] 					= true,
	[PublicEvent.PublicEventType_PVP_Battleground_Cannon] 		= true,
	[PublicEvent.PublicEventType_PVP_Battleground_Vortex] 		= true,
	[PublicEvent.PublicEventType_PVP_Battleground_Sabotage] 	= true,
	[PublicEvent.PublicEventType_PVP_Battleground_HoldTheLine] 	= true,
}

function LocalQuestTrackerAdvanced:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self
    return o
end

function LocalQuestTrackerAdvanced:Init()
    Apollo.RegisterAddon(self, true)
end

function LocalQuestTrackerAdvanced:OnLoad() -- OnLoad then GetAsyncLoad then OnRestore
	self.xmlDoc = XmlDoc.CreateFromFile("LocalQuestTrackerAdvanced.xml")
	self.xmlDocSettings = XmlDoc.CreateFromFile("Settings.xml")
	
	-- Set Settings to Default
	self.settings = {}
	for k, v in pairs(tDefaultSettings) do
		self.settings[k] = v
	end
	
	-- Gemini Support
	GeminiLogging = Apollo.GetPackage("Gemini:Logging-1.2").tPackage
	log = GeminiLogging:GetLogger({
  		level = GeminiLogging.INFO,
  		pattern = "%d %n %c %l - %m",
  		appender = "GeminiConsole"
	})
	log:info(string.format("Loaded LocalQuest Tracker Advanced %s", version))
end

function LocalQuestTrackerAdvanced:GetAsyncLoadStatus()
	if self.xmlDoc:IsLoaded() and self.xmlDocSettings:IsLoaded() and g_InterfaceOptionsLoaded then
		self:Initialize()
		return Apollo.AddonLoadStatus.Loaded
	end
	return Apollo.AddonLoadStatus.Loading
end

function LocalQuestTrackerAdvanced:Initialize()

	Apollo.RegisterEventHandler("OptionsUpdated_QuestTracker", 				"OnOptionsUpdated", self)
	Apollo.RegisterEventHandler("OptionsUpdated_LocalQuestTrackerAdvanced", "OnOptionsUpdated", self) -- Not sure if the Event is fired for original or this addon
	
	self.tMinimized =
	{
		tQuest = {},
		tEpisode = {},
		tQuestTime = {},
		tEpisodeTime = {},
	}

    self.tQuestTrackerBlink = ApolloTimer.Create(4, false, "OnQuestTrackerBlinkTimer", self)
	self.tQuestTrackerBlink:Stop()

	-- Code events, mostly to remove completed/finished quests
	-- TODO: an event needs to wndQuest:FindChild("ObjectiveContainer"):DestroyChildren() when moving to complete/botched
	Apollo.RegisterEventHandler("EpisodeStateChanged", 						"DestroyAndRedraw", self)
	Apollo.RegisterEventHandler("QuestStateChanged", 						"OnQuestStateChanged", self)
	Apollo.RegisterEventHandler("QuestObjectiveUpdated", 					"OnQuestObjectiveUpdated", self)
	Apollo.RegisterEventHandler("GenericEvent_QuestLog_TrackBtnClicked", 	"OnDestroyQuestObject", self) -- This is an event from QuestLog
	Apollo.RegisterEventHandler("Tutorial_RequestUIAnchor", 				"OnTutorial_RequestUIAnchor", self)
	Apollo.RegisterEventHandler("Communicator_ShowQuestMsg",				"OnShowCommMsg", self)
	Apollo.RegisterEventHandler("QuestInit",								"OnQuestInit", self)
	Apollo.RegisterEventHandler("SubZoneChanged",							"OnSubZoneChanged", self)

	-- Public Events
	Apollo.RegisterEventHandler("PublicEventEnd", 							"OnPublicEventEnd", self)
	Apollo.RegisterEventHandler("PublicEventLeave", 						"OnPublicEventEnd", self)
	Apollo.RegisterEventHandler("PublicEventStart", 						"OnPublicEventStart", self)
	Apollo.RegisterEventHandler("PublicEventObjectiveUpdate", 				"OnPublicEventUpdate", self)
	Apollo.RegisterEventHandler("PVPMatchFinished", 						"OnLeavePvP", self)
	Apollo.RegisterEventHandler("MatchExited", 								"OnLeavePvP", self)

	-- Formatting events
	Apollo.RegisterEventHandler("DatachronRestored", 						"SetPosition", self)
	Apollo.RegisterEventHandler("DatachronMinimized", 						"SetPosition", self)
	Apollo.RegisterEventHandler("GenericEvent_ChallengeTrackerToggled", 	"OnGenericEvent_ChallengeTrackerToggled", self)
	Apollo.RegisterEventHandler("QuestLog_ToggleLongQuestText", 			"OnToggleLongQuestText", self)

	-- Checking Player Death (can't turn in quests if dead)
	Apollo.RegisterEventHandler("PlayerResurrected", 						"OnPlayerResurrected", self)
	Apollo.RegisterEventHandler("ShowResurrectDialog", 						"OnShowResurrectDialog", self)

	Apollo.RegisterTimerHandler("OneSecTimer", 								"RedrawTimed", self)
	Apollo.RegisterTimerHandler("QuestTracker_EarliestProgBarTimer", 		"OnQuestTracker_EarliestProgBarTimer", self)
	
    self.tQuestTrackerRedraw = ApolloTimer.Create(0.2, false, "RedrawAll", self)
	
	-- LQTA Events --
	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)	

	Apollo.RegisterEventHandler("VarChange_ZoneName", 						"OnChangeZoneName", self)
	Apollo.RegisterEventHandler("SubZoneChanged", 							"OnChangeSubZoneName", self)
	
	Apollo.RegisterSlashCommand("lqta", "OnSlashCommand", self)
	
	Apollo.RegisterEventHandler("ResolutionChanged", "SetPosition", self)	
	Apollo.RegisterEventHandler("LQTA_ToggleOptions", "ShowHideSettings", self) -- Don't know id used anymore
	Apollo.RegisterTimerHandler("MainEscapableTimer", "OnMainEscapableTimer", self)
	Apollo.RegisterTimerHandler("LQTAFirstRunTimer", "OnFirstRun", self)
	Apollo.CreateTimer("LQTAFirstRunTimer", 4, false)
	
    self.tCheckCraftingStationDistance = ApolloTimer.Create(0.2, true, "OnCheckCraftingStationDistanceTimer", self)

    self.tDestroyAndRedraw = ApolloTimer.Create(0.2, false, "DestroyAndRedraw", self)
	self.tDestroyAndRedraw:Stop()

	-- Create windows --
	self.wndMain = Apollo.LoadForm(self.xmlDoc, "LocalQuestTrackerAdvancedForm", "FixedHudStratum", self)
	self.wndQuestTrackerScroll = self.wndMain:FindChild("QuestTrackerScroll")
	self.wndMain:SetSizingMinimum(325, 120)
	self.bMoveable = self.wndMain:IsStyleOn("Moveable")
	self.nLeft, self.nTop, self.nRight, self.nBottom = self.wndMain:GetAnchorOffsets()
	self:ApplyFreePositionSize()
	self:ApplyFilterPreferences()

	-- session variables --
	local unitPlayer = GameLib.GetPlayerUnit()
	self.bQuestTrackerByDistance 		= g_InterfaceOptions.Carbine.bQuestTrackerByDistance
	self.nQuestCounting 				= 0
	self.strPlayerPath 					= ""
	self.nFlashThisQuest 				= nil
	self.bPlayerIsDead 					= GameLib.GetPlayerUnit() and GameLib.GetPlayerUnit():IsDead() or false
	self.bDrawPvPScreenOnly 			= false
	self.bDrawDungeonScreenOnly 		= false
	self.tZombiePublicEvents 			= {}
	self.tActiveProgBarQuests 			= {}
	self.ZombieTimerMax					= 120 -- Time it takes for a zombie PE to dissapear
	self.tClickBlinkingQuest			= nil
	self.tHoverBlinkingQuest			= nil
	self.bMaximized						= false
	self.bRedrawQueued					= false
	self.tQuestsQueuedForDestroy		= {}
	self.tPublicEventsToRedraw			= {}
	self.tTimedQuests					= {}
	self.tTimedObjectives				= {}
	self.tTimedEvents					= {}
	self.tTimedEventObjectives			= {}
	self.tQueuedCommMessages			= {}
	self.bChallengeVisible				= nil -- TODO: Challenges is firing the toggle event for basically any event that happens regardless of whether or not it toggled. Once this is fixed, we can remove this state storage var.

	-- LQTA session variables--
	self.currentZone					= nil
	self.currentZoneId					= nil
	self.currentSubZone					= nil
	self.currentSubZoneId				= nil
	self.bOptionsOpen					= false
	self.bMainAnchorShown				= false
	self.showOnlyTradeskills			= false
	self.craftingStationList			= {}


	-- Draw Window --
	self:InitializeWindowMeasuring()
	self:RedrawAll()

	self:AutoTrack()
end

-- Custom Info for Chat
function LocalQuestTrackerAdvanced:OnFirstRun()
	if self.settings.bFirstRun then
		VILib:WriteToChat("[----------------]")
		VILib:WriteToChat("[Local Quest     ] Type /lqta to open the settings")
		VILib:WriteToChat("[Tracker Advanced] You can also access the them through")
		VILib:WriteToChat("[----------------] the label in the interface menu.")
	end
end

function LocalQuestTrackerAdvanced:InitializeWindowMeasuring() -- Try not to run these OnLoad as they may be expensive
	local wndMeasure = Apollo.LoadForm(self.xmlDoc, "EpisodeGroupItem", nil, self)
	self.knInitialEpisodeGroupHeight = wndMeasure:GetHeight()
	wndMeasure:Destroy()

	wndMeasure = Apollo.LoadForm(self.xmlDoc, "EpisodeItem", nil, self)
	self.knInitialEpisodeHeight = wndMeasure:GetHeight()
	wndMeasure:Destroy()

	wndMeasure = Apollo.LoadForm(self.xmlDoc, "QuestItem", nil, self)
	self.knInitialQuestControlBackerHeight = wndMeasure:FindChild("ControlBackerBtn"):GetHeight()
	wndMeasure:Destroy()

	wndMeasure = Apollo.LoadForm(self.xmlDoc, "QuestObjectiveItem", nil, self)
	self.knInitialQuestObjectiveHeight = wndMeasure:GetHeight()
	wndMeasure:Destroy()

	wndMeasure = Apollo.LoadForm(self.xmlDoc, "SpellItem", nil, self)
	self.knInitialSpellItemHeight = wndMeasure:GetHeight()
	wndMeasure:Destroy()

	wndMeasure = Apollo.LoadForm(self.xmlDoc, "EventItem", nil, self)
	self.knMinHeightEventItem = wndMeasure:GetHeight()
	wndMeasure:Destroy()

	if self.strPlayerPath == "" then
		local ePlayPathType = PlayerPathLib.GetPlayerPathType()
		if ePlayPathType then
			self.strPlayerPath = karPathToString[ePlayerPathType]
		end
	end
end

function LocalQuestTrackerAdvanced:OnOptionsUpdated()
	self.bQuestTrackerByDistance = g_InterfaceOptions.Carbine.bQuestTrackerByDistance
end

function LocalQuestTrackerAdvanced:OnQuestInit()
	self:AutoTrack()
end

function LocalQuestTrackerAdvanced:OnSubZoneChanged()
	self:RedrawAll()
	Apollo.RemoveEventHandler("SubZoneChanged", self) -- Only needed this event for the edge case of first load not having a zone set yet
end

function LocalQuestTrackerAdvanced:ResizeEpisodes()
	-- Sort
	local function HelperSortEpisodes(a,b)
		if a:FindChild("EpisodeTitle") and b:FindChild("EpisodeTitle") then
			return a:FindChild("EpisodeTitle"):GetData() < b:FindChild("EpisodeTitle"):GetData()
		elseif b:GetName() == "SwapToQuests" then
			return true
		end
		return false
	end

	for idx1, wndEpisodeGroup in pairs(self.wndQuestTrackerScroll:GetChildren()) do
		if wndEpisodeGroup:GetName() == "EpisodeGroupItem" then
			-- Resize List
			self:OnResizeEpisodeGroup(wndEpisodeGroup)
			wndEpisodeGroup:FindChild("EpisodeGroupContainer"):ArrangeChildrenVert(0, HelperSortEpisodes)
		elseif wndEpisodeGroup:GetName() == "EpisodeItem" then
			-- Resize List
			self:OnResizeEpisode(wndEpisodeGroup)
			wndEpisodeGroup:FindChild("EpisodeQuestContainer"):ArrangeChildrenVert(0, HelperSortEpisodes)
		end
	end
	
	-- local nAlign = self.bHasMoved and 0 or 2
	local nAlign = self.settings.bTopAlign and 0 or 2
	
	self.wndQuestTrackerScroll:ArrangeChildrenVert(nAlign, function(a,b)
		if a:GetName() == "EpisodeGroupItem" and b:GetName() == "EpisodeGroupItem" then
			return a:GetData() < b:GetData()
		elseif b:GetName() == "SwapToQuests" then
			return true
		end
		return false
	end)
end

function LocalQuestTrackerAdvanced:OnQuestTrackerBlinkTimer()
	self.tClickBlinkingQuest:SetActiveQuest(false)
	self.tClickBlinkingQuest = nil

	if self.tHoverBlinkingQuest then
		self.tHoverBlinkingQuest:ToggleActiveQuest()
	end
end

-----------------------------------------------------------------------------------------------
-- Main Redraw Methods
-----------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:QueueQuestForDestroy(queQuest)
	table.insert(self.tQuestsQueuedForDestroy, queQuest)
end

function LocalQuestTrackerAdvanced:RequestRedrawAll()
	if not self.bRedrawQueued then
		self.bRedrawQueued = true
		self.tQuestTrackerRedraw:Start()
	end
end

function LocalQuestTrackerAdvanced:DestroyAndRedraw()
	self.wndQuestTrackerScroll:DestroyChildren()
	self.tPublicEventsToRedraw = {}
	self.tTimedQuests = {}
	self.tTimedObjectives = {}
	self.tTimedEvents = {}
	self.tTimedEventObjectives = {}
	self:RedrawAll()
end

function LocalQuestTrackerAdvanced:RedrawTimed()

	if self.settings.bAutoCollapseEpisodes or self.settings.bAutoCollapseQuests then
		-- Clear unhide status on timed out quests/episodes
		if (self:UpdateTimeouts(self.tMinimized.tQuestTime, self.tMinimized.tQuest) or
			self:UpdateTimeouts(self.tMinimized.tEpisodeTime, self.tMinimized.tEpisode)) then
			self:RedrawAll()
		end
	end

	for index, tEventInfo in pairs(self.tTimedEvents) do
		if tEventInfo.peEvent:IsActive() and tEventInfo.wndTitleFrame and tEventInfo.wndTitleFrame:IsValid() then
			local strTitle = string.format("<T Font=\"CRB_InterfaceMedium\" TextColor=\"%s\">%s</T>", kstrLightGrey, tEventInfo.peEvent:GetName())
			strTitle = self:HelperPrefixTimeString(math.max(0, math.floor((tEventInfo.peEvent:GetTotalTime() - tEventInfo.peEvent:GetElapsedTime()) / 1000)), strTitle)
			tEventInfo.wndTitleFrame:SetAML(strTitle)
		else
			table.remove(self.tTimedEvents, index)
		end
	end
	
	for index, tEventObjectiveInfo in pairs(self.tTimedEventObjectives) do
		if tEventObjectiveInfo.peEvent:IsActive() and tEventObjectiveInfo.wndObjective and tEventObjectiveInfo.wndObjective:IsValid() and tEventObjectiveInfo.wndObjective:FindChild("QuestObjectiveBtn") ~= nil then
			tEventObjectiveInfo.wndObjective:FindChild("QuestObjectiveBtn"):SetTooltip(self:BuildEventObjectiveTitleString(tEventObjectiveInfo.peEvent, tEventObjectiveInfo.peoObjective, true))
			tEventObjectiveInfo.wndObjective:FindChild("QuestObjectiveText"):SetAML(self:BuildEventObjectiveTitleString(tEventObjectiveInfo.peEvent, tEventObjectiveInfo.peoObjective))
		else
			table.remove(self.tTimedEventObjectives, index)
		end
	end

	if not self.bDrawPvPScreenOnly and not self.bDrawDungeonScreenOnly then
		for index, tQuestInfo in pairs(self.tTimedQuests) do
			if tQuestInfo.wndTitleFrame ~= nil then
				local strTitle = self:HelperBuildTimedQuestTitle(tQuestInfo.queQuest)
				tQuestInfo.wndTitleFrame:SetAML(strTitle)
			else
				table.remove(self.tTimedQuests, index)
			end
		end
		
		for index, tObjectiveInfo in pairs(self.tTimedObjectives) do
			if tObjectiveInfo.wndObjective ~= nil and tObjectiveInfo.wndObjective:FindChild("QuestObjectiveBtn") ~= nil then
				tObjectiveInfo.wndObjective:FindChild("QuestObjectiveBtn"):SetTooltip(self:BuildObjectiveTitleString(tObjectiveInfo.queQuest, tObjectiveInfo.tObjective, true))
				tObjectiveInfo.wndObjective:FindChild("QuestObjectiveText"):SetAML(self:BuildObjectiveTitleString(tObjectiveInfo.queQuest, tObjectiveInfo.tObjective))
			else
				table.remove(self.tTimedObjectives, index)
			end
		end
	end
end

function LocalQuestTrackerAdvanced:RedrawAll()
	self.tQuestTrackerRedraw:Stop()
	self.bRedrawQueued = false
	
	self:HelperFindAndDestroyQuests()
	
	if #self.tZombiePublicEvents > 0 then
		self:DrawPublicEpisodes()
	elseif #PublicEvent.GetActiveEvents() > 0 then
		self:DrawPublicEpisodes()
	elseif self.wndQuestTrackerScroll:FindChildByUserData(kstrPublicEventMarker) then
		-- Safety (should rarely fire): If we're out of events and the window is still around, switch views.
		self.bDrawDungeonScreenOnly = false
		self.bDrawPvPScreenOnly = false
		self:DestroyAndRedraw()
		return
	end
	
	if not self.bDrawPvPScreenOnly and not self.bDrawDungeonScreenOnly then
		local wndEpisodeGroup
		local tCategoryHaveQuestsCache = {}
		local tCategoryEpisodeHaveQuestsCache = {}

		self.nQuestCounting = 0
		if false then
			-- Debugging
			for idx1, qcCategory in pairs(QuestLib.GetKnownCategories()) do
				if self:DoesCategoryHaveQuests(qcCategory, tCategoryHaveQuestsCache) then
					log:info(string.format("Category: "..qcCategory:GetTitle().."("..qcCategory:GetId()..")"))

					for idx2, epiEpisode in pairs(qcCategory:GetEpisodes()) do
						if not epiEpisode:IsWorldStory() and self:DoesCategoryEpisodeHaveQuests(qcCategory, epiEpisode, tCategoryEpisodeHaveQuestsCache) then
							log:info(string.format("   Episode: "..epiEpisode:GetTitle().."("..epiEpisode:GetId()..")"))

							for idx3, queQuest in pairs(epiEpisode:GetAllQuests(qcCategory:GetId())) do
								if (queQuest:GetState() ~= Quest.QuestState_Completed) and not queQuest:IsIgnored() then
									log:info(string.format("      Quest: "..queQuest:GetTitle().."("..queQuest:GetId()..")"))
								end
							end
						end
					end
				end
			end
		end
		if self.settings.bAutoShowHideTradeskills and self.showOnlyTradeskills then
			-- Tradeskills Quests
			for idx1, qcCategory in pairs(QuestLib.GetKnownCategories()) do
				if qcCategory:GetId() == 53 then -- Tradeskills Category
					for idx2, epiEpisode in pairs(qcCategory:GetEpisodes()) do
						if epiEpisode:GetId() == 1 then -- WorldTasks Episode (TODO Maybe there are Tradeskill Quests outside this Episode?)
							local wndTaskGroup = self:FactoryProduce(self.wndQuestTrackerScroll, "EpisodeGroupItem", "4EGTask")
							wndTaskGroup:FindChild("EpisodeGroupTitle"):SetText("Tradeskills")

							self:DrawEpisodeQuests(epiEpisode, wndTaskGroup:FindChild("EpisodeGroupContainer"))
						end
					end
				end
			end
		elseif self.settings.bAutoHideEpisodes then
			-- Auto Tracked Episodes
			local episodes = {}
			if self.currentZone == nil then
				self.currentZone = GameLib.GetCurrentZoneMap().strName
			end

			for idx1, qcCategory in pairs(QuestLib.GetKnownCategories()) do
				if self:DoesCategoryHaveQuests(qcCategory, tCategoryHaveQuestsCache) then
					for idx2, epiEpisode in pairs(qcCategory:GetEpisodes(self.bQuestTrackerByDistance)) do
						if not epiEpisode:IsWorldStory() and self:DoesCategoryEpisodeHaveQuests(qcCategory, epiEpisode, tCategoryEpisodeHaveQuestsCache) and (epiEpisode:GetZoneName() == "" or epiEpisode:GetZoneName() == self.currentZone) then
								episodes[epiEpisode:GetId()] = epiEpisode
						end
					end
				end
			end

			for idx, epiEpisode in pairs(episodes) do
				wndEpisodeGroup = nil
				if epiEpisode:IsWorldStory() then
					wndEpisodeGroup = self:FactoryProduce(self.wndQuestTrackerScroll, "EpisodeGroupItem", "1EGWorld")
					wndEpisodeGroup:FindChild("EpisodeGroupTitle"):SetText(Apollo.GetString("QuestTracker_WorldStory"))
				elseif epiEpisode:IsZoneStory() then
					wndEpisodeGroup = self:FactoryProduce(self.wndQuestTrackerScroll, "EpisodeGroupItem", "2EGZone")
					wndEpisodeGroup:FindChild("EpisodeGroupTitle"):SetText(Apollo.GetString("QuestTracker_ZoneStory"))
				elseif epiEpisode:IsRegionalStory() then
					wndEpisodeGroup = self:FactoryProduce(self.wndQuestTrackerScroll, "EpisodeGroupItem", "3EGRegional")
					wndEpisodeGroup:FindChild("EpisodeGroupTitle"):SetText(Apollo.GetString("QuestTracker_RegionalStory"))
				else -- task
					local wndTaskGroup = self:FactoryProduce(self.wndQuestTrackerScroll, "EpisodeGroupItem", "4EGTask")
					wndTaskGroup:FindChild("EpisodeGroupTitle"):SetText(Apollo.GetString("QuestTracker_Tasks"))

					self:DrawEpisodeQuests(epiEpisode, wndTaskGroup:FindChild("EpisodeGroupContainer"))
				end

				if wndEpisodeGroup ~= nil then
					self:DrawEpisode(idx, epiEpisode, wndEpisodeGroup:FindChild("EpisodeGroupContainer"))
				end
			end
		else
			-- Normally Tracked Episodes
			for idx, epiEpisode in pairs(QuestLib.GetTrackedEpisodes(self.bQuestTrackerByDistance)) do
				wndEpisodeGroup = nil
				if epiEpisode:IsWorldStory() then
					wndEpisodeGroup = self:FactoryProduce(self.wndQuestTrackerScroll, "EpisodeGroupItem", "1EGWorld")
					wndEpisodeGroup:FindChild("EpisodeGroupTitle"):SetText(Apollo.GetString("QuestTracker_WorldStory"))
				elseif epiEpisode:IsZoneStory() then
					wndEpisodeGroup = self:FactoryProduce(self.wndQuestTrackerScroll, "EpisodeGroupItem", "2EGZone")
					wndEpisodeGroup:FindChild("EpisodeGroupTitle"):SetText(Apollo.GetString("QuestTracker_ZoneStory"))
				elseif epiEpisode:IsRegionalStory() then
					wndEpisodeGroup = self:FactoryProduce(self.wndQuestTrackerScroll, "EpisodeGroupItem", "3EGRegional")
					wndEpisodeGroup:FindChild("EpisodeGroupTitle"):SetText(Apollo.GetString("QuestTracker_RegionalStory"))
				else -- task
					local wndTaskGroup = self:FactoryProduce(self.wndQuestTrackerScroll, "EpisodeGroupItem", "4EGTask")
					wndTaskGroup:FindChild("EpisodeGroupTitle"):SetText(Apollo.GetString("QuestTracker_Tasks"))

					self:DrawEpisodeQuests(epiEpisode, wndTaskGroup:FindChild("EpisodeGroupContainer"))
				end

				if wndEpisodeGroup ~= nil then
					self:DrawEpisode(idx, epiEpisode, wndEpisodeGroup:FindChild("EpisodeGroupContainer"))
				end
			end
		end

		wndEpisodeGroup = self.wndQuestTrackerScroll:FindChildByUserData("1EGWorld")
		if wndEpisodeGroup ~= nil and wndEpisodeGroup:IsValid() and next(wndEpisodeGroup:FindChild("EpisodeGroupContainer"):GetChildren()) == nil then
			wndEpisodeGroup:Destroy()
		end
		wndEpisodeGroup = self.wndQuestTrackerScroll:FindChildByUserData("2EGZone")
		if wndEpisodeGroup ~= nil and wndEpisodeGroup:IsValid() and next(wndEpisodeGroup:FindChild("EpisodeGroupContainer"):GetChildren()) == nil then
			wndEpisodeGroup:Destroy()
		end
		wndEpisodeGroup = self.wndQuestTrackerScroll:FindChildByUserData("3EGRegional")
		if wndEpisodeGroup ~= nil and wndEpisodeGroup:IsValid() and next(wndEpisodeGroup:FindChild("EpisodeGroupContainer"):GetChildren()) == nil then
			wndEpisodeGroup:Destroy()
		end
		wndEpisodeGroup = self.wndQuestTrackerScroll:FindChildByUserData("4EGTask")
		if wndEpisodeGroup ~= nil and wndEpisodeGroup:IsValid() and next(wndEpisodeGroup:FindChild("EpisodeGroupContainer"):GetChildren()) == nil then
			wndEpisodeGroup:Destroy()
		end
	end
	
	self:ResizeEpisodes()
end

function LocalQuestTrackerAdvanced:DrawEpisode(idx, epiEpisode, wndParent)
	local wndEpisode = self:FactoryProduce(wndParent, "EpisodeItem", epiEpisode)
	wndEpisode:FindChild("EpisodeTitle"):SetData(idx) -- For sorting
	wndEpisode:FindChild("EpisodeMinimizeBtn"):SetData(epiEpisode:GetId())
	
	-- Auto Collapse Episodes that are out of your current area --
	if self.currentZone == nil then
		self.currentZone = GameLib.GetCurrentZoneMap().strName
	end
	if self.settings.bAutoCollapseEpisodes then
		local hidden = true
		
		if (epiEpisode:GetZoneName() ~= "" and epiEpisode:GetZoneName() == self.currentZone) or (self.tMinimized.tEpisodeTime[epiEpisode:GetId()] ~= nil and (os.time() - self.tMinimized.tEpisodeTime[epiEpisode:GetId()]) < self.settings.nTimeout) then
			hidden = false
		end

		local unhide = self.tMinimized.tEpisode[epiEpisode:GetId()]
		if (unhide ~= nil) then
			hidden = unhide
		end
		
		wndEpisode:FindChild("EpisodeMinimizeBtn"):SetCheck(hidden)
	else
		if self.tMinimized.tEpisode[epiEpisode:GetId()] then
			wndEpisode:FindChild("EpisodeMinimizeBtn"):SetCheck(true)
		end
	end
	


	if wndEpisode:FindChild("EpisodeMinimizeBtn") and wndEpisode:FindChild("EpisodeMinimizeBtn"):IsChecked() then
		wndEpisode:FindChild("EpisodeTitle"):SetText("> " .. epiEpisode:GetTitle())
		wndEpisode:FindChild("EpisodeTitle"):SetTextColor(ApolloColor.new("8031fcf6"))

		-- Flash if we are told to
		if self.nFlashThisQuest then
			for key, queQuest in pairs(epiEpisode:GetTrackedQuests()) do
				self.nQuestCounting = self.nQuestCounting + 1
				if self.nFlashThisQuest == queQuest then
					self.nFlashThisQuest = nil
					wndEpisode:FindChild("EpisodeTitle"):SetSprite("sprTrk_ObjectiveUpdatedAnim")
				end
			end
		else
			for key, queQuest in pairs(epiEpisode:GetTrackedQuests()) do
				self.nQuestCounting = self.nQuestCounting + 1
			end
		end
	elseif wndEpisode:FindChild("EpisodeMinimizeBtn") then
		wndEpisode:FindChild("EpisodeTitle"):SetText(epiEpisode:GetTitle())
		wndEpisode:FindChild("EpisodeTitle"):SetTextColor(ApolloColor.new("UI_BtnTextHoloNormal"))

		self:DrawEpisodeQuests(epiEpisode, wndEpisode:FindChild("EpisodeQuestContainer"))
	end
end

function LocalQuestTrackerAdvanced:DrawEpisodeQuests(epiEpisode, wndContainer)
	if self.settings.bAutoShowHideTradeskills and self.showOnlyTradeskills then
		-- Auto Tracking Tradeskills
		for idx, queQuest in pairs(epiEpisode:GetAllQuests(53)) do
			if (queQuest:GetState() ~= Quest.QuestState_Completed) and not queQuest:IsIgnored() then
				self.nQuestCounting = self.nQuestCounting + 1
				self:DrawQuest(self.nQuestCounting, queQuest, wndContainer)
			end
		end
	else
		-- Normally Tracked Quests
		for nIdx, queQuest in pairs(epiEpisode:GetTrackedQuests(0, self.bQuestTrackerByDistance)) do
			self.nQuestCounting = self.nQuestCounting + 1
			self:DrawQuest(self.nQuestCounting, queQuest, wndContainer)
		end
	end

	-- Inline Sort Method
	local function SortQuestTrackerScroll(a, b)
	if not a or not b or not a:FindChild("QuestNumber") or not b:FindChild("QuestNumber") then return true end
		return (tonumber(a:FindChild("QuestNumber"):GetText()) or 0) < (tonumber(b:FindChild("QuestNumber"):GetText()) or 0)
	end

	wndContainer:ArrangeChildrenVert(0, SortQuestTrackerScroll)
end

function LocalQuestTrackerAdvanced:DrawQuest(nIdx, queQuest, wndParent)
	local wndQuest = self:FactoryProduce(wndParent, "QuestItem", queQuest)
	
	-- Quest Title	
	local strTitle = queQuest:GetTitle()
	local eQuestState = queQuest:GetState()
	if eQuestState == Quest.QuestState_Botched then
		strTitle = string.format("<T Font=\"CRB_InterfaceMedium_B\" TextColor=\"%s\">%s</T>", kstrRed, String_GetWeaselString(Apollo.GetString("QuestTracker_Failed"), strTitle))
	elseif eQuestState == Quest.QuestState_Achieved then
		strTitle = string.format("<T Font=\"CRB_InterfaceMedium_B\" TextColor=\"%s\">%s</T>", kstrGreen,String_GetWeaselString(Apollo.GetString("QuestTracker_Complete"), strTitle))
	elseif (eQuestState == Quest.QuestState_Accepted or eQuestState == Quest.QuestState_Achieved) and queQuest:IsQuestTimed() then
		strTitle = self:HelperBuildTimedQuestTitle(queQuest)
		table.insert(self.tTimedQuests, { queQuest = queQuest, wndTitleFrame = wndQuest:FindChild("TitleText") })
	else
		local strColor = self.tActiveProgBarQuests[queQuest:GetId()] and "ffffffff" or kstrLightGrey
		local crLevelConDiff = ktConToColor[queQuest:GetColoredDifficulty() or 0]
		strTitle = string.format("<T Font=\"CRB_InterfaceMedium\" TextColor=\"%s\">%s </T><T Font=\"CRB_InterfaceMedium\" TextColor=\"%s\">[%s]</T>", strColor, strTitle, crLevelConDiff, queQuest:GetConLevel())
	end
	
	wndQuest:FindChild("TitleText"):SetAML(strTitle)
	wndQuest:FindChild("TitleText"):SetHeightToContentHeight()
		
	wndQuest:FindChild("QuestOpenLogBtn"):SetData(queQuest)
	wndQuest:FindChild("QuestCallbackBtn"):SetData(queQuest)
	wndQuest:FindChild("ControlBackerBtn"):SetData(wndQuest)
	wndQuest:FindChild("QuestCloseBtn"):SetData({["wndQuest"] = wndQuest, ["tQuest"] = queQuest})
	wndQuest:FindChild("MinimizeBtn"):SetData(queQuest:GetId())

	-- Flash if we are told to
	if self.nFlashThisQuest == queQuest then
		self.nFlashThisQuest = nil
		wndQuest:SetSprite("sprWinAnim_BirthSmallTemp")
	end

	-- Auto Collapse Quests that are out of your current region --
	if self.settings.bAutoCollapseQuests and not self.showOnlyTradeskills then
		--local zoneName = queQuest:GetEpisode():GetZoneName()
		local regions = queQuest:GetMapRegions()	
		local hidden = true

		if regions ~= 0 then
			for idx,questRegion in pairs(regions) do
				local questZoneId = questRegion.nWorldZoneId
				if (GameLib.IsInWorldZone(questZoneId) == true or GameLib.IsZoneInZone(GameLib.GetCurrentZoneId(), questZoneId) == true) then
					hidden = false
				end
			end
		end

		if (self.tMinimized.tQuestTime[queQuest:GetId()] ~= nil and (os.time() - self.tMinimized.tQuestTime[queQuest:GetId()]) < self.settings.nTimeout) then
			hidden = false
		end
			
		local unhide = self.tMinimized.tQuest[queQuest:GetId()]
		if  unhide ~= nil then
			hidden = unhide
		end
		
		wndQuest:FindChild("MinimizeBtn"):SetCheck(hidden)
	else
		if queQuest:GetId() and self.tMinimized.tQuest[queQuest:GetId()] then
			wndQuest:FindChild("MinimizeBtn"):SetCheck(true)
		end
	end



	-- Quest spell
	if queQuest:GetSpell() then
		local wndSpellItem = self:FactoryProduce(wndQuest:FindChild("ObjectiveContainer"), "SpellItem", "SpellItem")
		wndSpellItem:FindChild("SpellItemBtn"):Show(true)
		wndSpellItem:FindChild("SpellItemBtn"):SetContentId(queQuest) -- GOTCHA: Normally we use the spell id, but here we use the quest object
		wndSpellItem:FindChild("SpellItemText"):SetText(String_GetWeaselString(Apollo.GetString("QuestTracker_UseQuestAbility"), GameLib.GetKeyBinding("CastObjectiveAbility")))
	end

	-- Conditional drawing
	wndQuest:FindChild("QuestNumberUpdateHighlight"):Show(self.tActiveProgBarQuests[queQuest:GetId()] ~= nil)
	wndQuest:FindChild("QuestNumber"):SetText(nIdx)
	wndQuest:FindChild("QuestNumber"):SetTextColor(ApolloColor.new("ff31fcf6"))
	wndQuest:FindChild("QuestCompletedBacker"):Show(false)
	wndQuest:FindChild("QuestNumberBackerArt"):SetBGColor(CColor.new(1,1,1,1))
	wndQuest:FindChild("QuestNumberBackerArt"):SetSprite("sprQT_NumBackerNormal")
	wndQuest:FindChild("ObjectiveContainer"):Show(not wndQuest:FindChild("MinimizeBtn"):IsChecked())
	
	-- State depending drawing
	if wndQuest:FindChild("MinimizeBtn"):IsChecked() then
		wndQuest:FindChild("QuestNumber"):SetTextColor(CColor.new(.5, .5, .5, .8))
		wndQuest:FindChild("QuestNumberBackerArt"):SetBGColor(CColor.new(.5, .5, .5, .8))

	elseif eQuestState == Quest.QuestState_Botched then
		self:HelperShowQuestCallbackBtn(wndQuest, queQuest, "sprQT_NumBackerFailed", "CRB_QuestTrackerSprites:btnQT_QuestFailed")
		wndQuest:FindChild("QuestNumber"):SetTextColor(ApolloColor.new(kstrRed))

	elseif eQuestState == Quest.QuestState_Achieved then
		self:HelperShowQuestCallbackBtn(wndQuest, queQuest, "sprQT_NumBackerCompleted", "CRB_QuestTrackerSprites:btnQT_QuestRedeem")
		wndQuest:FindChild("QuestNumber"):SetTextColor(ApolloColor.new("ff7fffb9"))

		-- Achieve objective only has one
		local wndObjective = self:FactoryProduce(wndQuest:FindChild("ObjectiveContainer"), "QuestObjectiveItem", "ObjectiveCompleted")
		wndObjective:FindChild("QuestObjectiveBtn"):SetTooltip(self:BuildObjectiveTitleString(queQuest, tObjective, true))
		wndObjective:FindChild("QuestObjectiveBtn"):SetData({["queOwner"] = queQuest, ["nObjectiveIdx"] = nil})
		wndObjective:FindChild("QuestObjectiveText"):SetAML(self:BuildObjectiveTitleString(queQuest))

	else
		-- Potentially multiple objectives if not minimized or in the achieved/botched state
		for idObjective, tObjective in pairs(queQuest:GetVisibleObjectiveData()) do
			if tObjective.nCompleted < tObjective.nNeeded then
				local wndObjective = self:FactoryProduce(wndQuest:FindChild("ObjectiveContainer"), "QuestObjectiveItem", idObjective)
				self:DrawQuestObjective(wndQuest, wndObjective, queQuest, tObjective)
				
				if queQuest:IsObjectiveTimed(tObjective.nIndex) then
					table.insert(self.tTimedObjectives, { queQuest = queQuest, tObjective = tObjective, wndObjective = wndObjective })
				end
			end
		end
	end

	wndQuest:FindChild("ObjectiveContainer"):ArrangeChildrenVert(0)
end

function LocalQuestTrackerAdvanced:DrawQuestObjective(wndQuest, wndObjective, queQuest, tObjective)
	wndObjective:FindChild("QuestObjectiveBtn"):SetData({["queOwner"] = queQuest, ["nObjectiveIdx"] = tObjective.nIndex})
	wndObjective:FindChild("QuestObjectiveBtn"):SetTooltip(self:BuildObjectiveTitleString(queQuest, tObjective, true))
	wndObjective:FindChild("QuestObjectiveText"):SetAML(self:BuildObjectiveTitleString(queQuest, tObjective))

	-- Progress
	if self.tActiveProgBarQuests[queQuest:GetId()] and queQuest:DisplayObjectiveProgressBar(tObjective.nIndex) then
		local wndObjectiveProg = self:FactoryProduce(wndObjective, "QuestProgressItem", "QuestProgressItem")
		local nCompleted = tObjective.nCompleted
		local nNeeded = tObjective.nNeeded
		wndObjectiveProg:FindChild("QuestProgressBar"):SetMax(nNeeded)
		wndObjectiveProg:FindChild("QuestProgressBar"):SetProgress(nCompleted)
		wndObjectiveProg:FindChild("QuestProgressBar"):EnableGlow(nCompleted > 0 and nCompleted ~= nNeeded)
	elseif wndObjective:FindChild("QuestProgressItem") then
		wndObjective:FindChild("QuestProgressItem"):Destroy()
		self:RedrawAll() -- TODO: this sucks, we trigger a redraw all while we're in the middle of already redrawing all
	end

	-- Objective Spell Item
	if queQuest:GetSpell(tObjective.nIndex) then
		local wndSpellBtn = self:FactoryProduce(wndObjective, "SpellItemObjectiveBtn", "SpellItemObjectiveBtn"..tObjective.nIndex)
		wndSpellBtn:SetContentId(queQuest, tObjective.nIndex)
	end
end

function LocalQuestTrackerAdvanced:DrawPublicEpisodes()
	local tPublicEvents = PublicEvent.GetActiveEvents()
	if self.bDrawPvPScreenOnly or self.bDrawDungeonScreenOnly then
		self:FactoryProduce(self.wndQuestTrackerScroll, "SwapToQuests", "SwapToQuests")
	elseif not self.wndMain:FindChild("SwapToPvP") and not self.wndMain:FindChild("SwapToDungeons") then
		for key, peEvent in pairs(tPublicEvents) do
			if not self.bDrawPvPScreenOnly and ktPvPEventTypes[peEvent:GetEventType()] then
				self.bDrawPvPScreenOnly = true
				self:DestroyAndRedraw()
				self:FactoryProduce(self.wndQuestTrackerScroll, "SwapToQuests", "SwapToQuests")
				return
			end
			if not self.bDrawDungeonScreenOnly and peEvent:GetEventType() == PublicEvent.PublicEventType_Dungeon then
				self.bDrawDungeonScreenOnly = true
				self:DestroyAndRedraw()
				self:FactoryProduce(self.wndQuestTrackerScroll, "SwapToQuests", "SwapToQuests")
				return
			end
		end
	end

	local wndEpisode = self:FactoryProduce(self.wndQuestTrackerScroll, "EpisodeItem", kstrPublicEventMarker)
	wndEpisode:FindChild("EpisodeTitle"):SetData(-1) -- For sorting, will compare vs Quests

	if wndEpisode:FindChild("EpisodeMinimizeBtn") and wndEpisode:FindChild("EpisodeMinimizeBtn"):IsChecked() then
		wndEpisode:FindChild("EpisodeTitle"):SetText("> " .. Apollo.GetString("QuestTracker_Events"))
		wndEpisode:FindChild("EpisodeTitle"):SetTextColor(ApolloColor.new("8031fcf6"))
		return
	end

	wndEpisode:FindChild("EpisodeTitle"):SetText(Apollo.GetString("QuestTracker_Events"))
	wndEpisode:FindChild("EpisodeTitle"):SetTextColor(ApolloColor.new("UI_BtnTextHoloNormal"))

	-- Events
	local nAlphabetNumber = 0
	for key, peEvent in pairs(tPublicEvents) do
		nAlphabetNumber	= nAlphabetNumber + 1
		self:DrawEvent(wndEpisode:FindChild("EpisodeQuestContainer"), peEvent, nAlphabetNumber)
	end

	-- Trim zombies to max size
	if #self.tZombiePublicEvents > knMaxZombieEventCount then
		table.remove(self.tZombiePublicEvents, 1)
	end

	-- Check Zombie Timer
	for key, tZombieEvent in pairs(self.tZombiePublicEvents) do
		tZombieEvent["nTimer"] = tZombieEvent["nTimer"] - 1
		if tZombieEvent["nTimer"] <= 0 then
			table.remove(self.tZombiePublicEvents, key)
			self:DestroyAndRedraw()
			return
		end
	end

	-- Now Draw Completed Events
	for key, tZombieEvent in pairs(self.tZombiePublicEvents) do
		nAlphabetNumber	= nAlphabetNumber + 1
		self:DrawZombieEvent(wndEpisode:FindChild("EpisodeQuestContainer"), tZombieEvent, nAlphabetNumber)
	end

	-- Inline Sort Method
	local function SortEventTrackerScroll(a, b)
		if not Window.is(a) or not Window.is(b) or not a:IsValid() or not b:IsValid() then
			return false
		end
		return a:FindChild("EventLetter"):GetText() < b:FindChild("EventLetter"):GetText()
	end

	wndEpisode:FindChild("EpisodeQuestContainer"):ArrangeChildrenVert(0, SortEventTrackerScroll)
end

function LocalQuestTrackerAdvanced:DrawEvent(wndParent, peEvent, nAlphabetNumber)
	local wndEvent = self:FactoryProduce(wndParent, "EventItem", peEvent)
	
	if wndEvent:FindChild("ShowEventStatsBtn"):GetData() and not self.tPublicEventsToRedraw[peEvent:GetName()] then
		return
	end
	
	self.tPublicEventsToRedraw[peEvent:GetName()] = false
	wndEvent:FindChild("ShowEventStatsBtn"):SetData(peEvent)
	wndEvent:FindChild("QuestMouseCatcher"):SetData(wndEvent)

	-- Event Title
	local strTitle = string.format("<T Font=\"CRB_InterfaceMedium\" TextColor=\"%s\">%s</T>", kstrLightGrey, peEvent:GetName())
	if peEvent:GetTotalTime() > 0 and peEvent:IsActive() then
		strTitle = self:HelperPrefixTimeString(math.max(0, math.floor((peEvent:GetTotalTime() - peEvent:GetElapsedTime()) / 1000)), strTitle)
		table.insert(self.tTimedEvents, { peEvent = peEvent, wndTitleFrame = wndEvent:FindChild("TitleText") })
	end
	wndEvent:FindChild("TitleText"):SetAML(strTitle)
	wndEvent:FindChild("TitleText"):SetHeightToContentHeight()

	-- Conditional Drawing
	wndEvent:FindChild("EventStatsBacker"):Show(peEvent:HasLiveStats())
	wndEvent:FindChild("EventLetter"):SetText(ktNumbersToLetters[nAlphabetNumber])
	wndEvent:FindChild("EventLetter"):SetTextColor(ApolloColor.new("ff31fcf6"))
	wndEvent:FindChild("EventLetterBacker"):SetBGColor(CColor.new(1,1,1,1))

	if wndEvent:FindChild("MinimizeBtn"):IsChecked() then
		wndEvent:FindChild("EventLetter"):SetTextColor(CColor.new(.5, .5, .5, .8))
		wndEvent:FindChild("EventLetterBacker"):SetBGColor(CColor.new(.5, .5, .5, .8))
	else
		-- Draw the Objective, or delete if it's still around
		for idObjective, peoObjective in pairs(peEvent:GetObjectives()) do
			if peoObjective:GetStatus() == PublicEventObjective.PublicEventStatus_Active and not peoObjective:IsHidden() then
				local wndObjective = self:FactoryProduce(wndEvent:FindChild("ObjectiveContainer"), "QuestObjectiveItem", peoObjective)
				self:DrawEventObjective(wndObjective, peEvent, idObjective, peoObjective)
			elseif wndEvent:FindChild("ObjectiveContainer"):FindChildByUserData(peoObjective) then
				wndEvent:FindChild("ObjectiveContainer"):FindChildByUserData(peoObjective):Destroy()
			end
		end

		-- Inline Sort Method
		local function SortEventObjectivesTrackerScroll(a, b)
			if not Window.is(a) or not Window.is(b) or not a:IsValid() or not b:IsValid() or not a:GetData() or not b:GetData() then
				return false
			end
			return a:GetData():GetCategory() < b:GetData():GetCategory()
		end

		wndEvent:FindChild("ObjectiveContainer"):ArrangeChildrenVert(0, SortEventObjectivesTrackerScroll)
	end
end

function LocalQuestTrackerAdvanced:DrawZombieEvent(wndParent, tZombieEvent, nAlphabetNumber)
	local wndEvent = self:FactoryProduce(wndParent, "ZombieEventItem", tZombieEvent.peEvent)
	wndEvent:FindChild("QuestCallbackBtn"):SetData(wndEvent)
	wndEvent:FindChild("QuestMouseCatcher"):SetData(wndEvent)

	-- Conditional Drawing
	wndEvent:FindChild("EventLetter"):SetText(ktNumbersToLetters[nAlphabetNumber])
	wndEvent:FindChild("EventLetterBacker"):SetBGColor("white")

	if wndEvent:FindChild("MinimizeBtn"):IsChecked() then
		wndEvent:FindChild("EventLetter"):SetTextColor(CColor.new(.5, .5, .5, .8))
		wndEvent:FindChild("EventLetterBacker"):SetBGColor(CColor.new(.5, .5, .5, .8))
	end

	-- Win or Loss formatting here
	local strTitle = string.format("<T Font=\"CRB_InterfaceMedium_B\">%s</T>", tZombieEvent.peEvent:GetName())
	if tZombieEvent.eReason == PublicEvent.PublicEventParticipantRemoveReason_CompleteFailure then
		local strFailed = String_GetWeaselString(Apollo.GetString("QuestTracker_Failed"), strTitle)
		wndEvent:FindChild("EventLetter"):SetTextColor(ApolloColor.new(kstrRed))
		wndEvent:FindChild("EventLetterBacker"):SetSprite("sprQT_NumBackerFailedPE")
		wndEvent:FindChild("QuestCallbackBtn"):ChangeArt("CRB_QuestTrackerSprites:btnQT_QuestFailed")
		wndEvent:FindChild("TitleText"):SetAML(string.format("<T Font=\"CRB_InterfaceMedium_B\" TextColor=\"%s\">%s</T>", kstrRed, strFailed))

	elseif tZombieEvent.eReason == PublicEvent.PublicEventParticipantRemoveReason_CompleteSuccess then
		local strComplete = String_GetWeaselString(Apollo.GetString("QuestTracker_Complete"), strTitle)
		wndEvent:FindChild("EventLetter"):SetTextColor(ApolloColor.new(kstrGreen))
		wndEvent:FindChild("EventLetterBacker"):SetSprite("sprQT_NumBackerCompletedPE")
		wndEvent:FindChild("QuestCallbackBtn"):ChangeArt("CRB_QuestTrackerSprites:btnQT_QuestRedeem")
		wndEvent:FindChild("TitleText"):SetAML(string.format("<T Font=\"CRB_InterfaceMedium_B\" TextColor=\"%s\">%s</T>", kstrGreen, strComplete))
	end
end

function LocalQuestTrackerAdvanced:DrawEventObjective(wndObjective, peEvent, idObjective, peoObjective)
	wndObjective:FindChild("QuestObjectiveBtn"):SetData({["peoObjective"] = peoObjective })
	wndObjective:FindChild("QuestObjectiveBtn"):SetTooltip(self:BuildEventObjectiveTitleString(peEvent, peoObjective, true))
	wndObjective:FindChild("QuestObjectiveText"):SetAML(self:BuildEventObjectiveTitleString(peEvent, peoObjective))

	if peoObjective:GetTotalTime() > 0 then
		table.insert(self.tTimedEventObjectives, { peEvent = peEvent, peoObjective = peoObjective, wndObjective = wndObjective })
	end
	
	-- Progress Bar
	if peoObjective:GetObjectiveType() == PublicEventObjective.PublicEventObjectiveType_ContestedArea then
		local nPercent = peoObjective:GetContestedAreaRatio()
		if peoObjective:GetContestedAreaOwningTeam() == 0 then
			nPercent = (nPercent + 100.0) * 0.5
		end

		local wndObjectiveProg = self:FactoryProduce(wndObjective, "PublicProgressItem", "PublicProgressItem")
		wndObjectiveProg:FindChild("PublicProgressBar"):SetMax(100)
		wndObjectiveProg:FindChild("PublicProgressBar"):SetProgress(nPercent)
		wndObjectiveProg:FindChild("PublicProgressBar"):EnableGlow(false)
		wndObjectiveProg:FindChild("PublicProgressText"):SetText(String_GetWeaselString(Apollo.GetString("CRB_Percent"), math.floor(nPercent)))

	elseif peoObjective:ShowPercent() or peoObjective:ShowHealthBar() then
		local wndObjectiveProg = self:FactoryProduce(wndObjective, "PublicProgressItem", "PublicProgressItem")
		local nCompleted = peoObjective:GetCount()
		local nNeeded = peoObjective:GetRequiredCount()
		wndObjectiveProg:FindChild("PublicProgressBar"):SetMax(nNeeded)
		wndObjectiveProg:FindChild("PublicProgressBar"):SetProgress(nCompleted)
		wndObjectiveProg:FindChild("PublicProgressBar"):EnableGlow(nCompleted > 0 and nCompleted ~= nNeeded)
		wndObjectiveProg:FindChild("PublicProgressText"):SetText(String_GetWeaselString(Apollo.GetString("CRB_Percent"), math.floor(nCompleted / nNeeded * 100)))
	end

	-- Objective Spell Item
	if peoObjective:GetSpell() then
		local wndSpellBtn = self:FactoryProduce(wndObjective, "SpellItemObjectiveBtn", idObjective)
		wndSpellBtn:SetContentId(peoObjective)
	end
end

function LocalQuestTrackerAdvanced:OnShowCommMsg(idMsg, idCaller, queUpdated, strText)
	local tCommInfo = self.tQueuedCommMessages and self.tQueuedCommMessages[queUpdated:GetId()] or nil
	if tCommInfo then
		self:HelperShowQuestCallbackBtn(tCommInfo.wndQuest, tCommInfo.queQuest, tCommInfo.strNumberBackerArt, tCommInfo.strCallbackBtnArt)
	end
end

-----------------------------------------------------------------------------------------------
-- Main Resize Method
-----------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:OnResizeEpisodeGroup(wndEpisodeGroup)
	local nOngoingGroupCount = self.knInitialEpisodeGroupHeight
	local wndEpisodeGroupContainer = wndEpisodeGroup:FindChild("EpisodeGroupContainer")
	
	for idx, wndEpisode in pairs(wndEpisodeGroupContainer:GetChildren()) do
		local strWindowName = wndEpisode:GetName()
		if strWindowName == "EpisodeItem" then
			nOngoingGroupCount = nOngoingGroupCount + self:OnResizeEpisode(wndEpisode)
		elseif strWindowName == "QuestItem" then
			nOngoingGroupCount = nOngoingGroupCount + self:OnResizeQuest(wndEpisode)
		end
	end
	
	wndEpisodeGroupContainer:ArrangeChildrenVert(0)
	local nLeft, nTop, nRight, nBottom = wndEpisodeGroup:GetAnchorOffsets()
	wndEpisodeGroup:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nOngoingGroupCount)
end

function LocalQuestTrackerAdvanced:OnResizeEpisode(wndEpisode)
	local nOngoingTopCount = self.knInitialEpisodeHeight
	local wndEpisodeQuestContainer = wndEpisode:FindChild("EpisodeQuestContainer")
	local bEpisodeMinimizeBtnChecked = wndEpisode:FindChild("EpisodeMinimizeBtn"):IsChecked()
	
	if not bEpisodeMinimizeBtnChecked then
		for idx1, wndQuest in pairs(wndEpisodeQuestContainer:GetChildren()) do
			nOngoingTopCount = nOngoingTopCount + self:OnResizeQuest(wndQuest)
		end
	end
	
	wndEpisodeQuestContainer:ArrangeChildrenVert(0)
	local nLeft, nTop, nRight, nBottom = wndEpisode:GetAnchorOffsets()
	wndEpisode:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nOngoingTopCount)
	wndEpisodeQuestContainer:Show(not bEpisodeMinimizeBtnChecked)
	return nOngoingTopCount
end

function LocalQuestTrackerAdvanced:OnResizeQuest(wndQuest)
	local nQuestTextWidth, nQuestTextHeight = wndQuest:FindChild("TitleText"):SetHeightToContentHeight()
	local nResult = math.max(self.knInitialQuestControlBackerHeight, nQuestTextHeight + 4) -- for lower g height
	
	local wndControlBackerBtn = wndQuest:FindChild("ControlBackerBtn")
	if wndControlBackerBtn then
		local nLeft, nTop, nRight, nBottom = wndControlBackerBtn:GetAnchorOffsets()
		wndControlBackerBtn:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nResult)
	end
	
	local wndObjectiveContainer = wndQuest:FindChild("ObjectiveContainer")
	local wndMinimizeBtn = wndQuest:FindChild("MinimizeBtn")
	
	local nHeaderHeight = nResult
	
	-- If expanded and valid, make it bigger
	if wndObjectiveContainer then
		if not wndMinimizeBtn:IsChecked() then
			for idx, wndObj in pairs(wndObjectiveContainer:GetChildren()) do
				nResult = nResult + self:OnResizeQuestObjective(wndObj)
			end
			
			local nLeft, nTop, nRight, nBottom = wndObjectiveContainer:GetAnchorOffsets()
			wndObjectiveContainer:SetAnchorOffsets(nLeft, nHeaderHeight, nRight, nHeaderHeight + wndObjectiveContainer:ArrangeChildrenVert(0))
		end
	
		wndObjectiveContainer:Show(not wndMinimizeBtn:IsChecked())
		wndObjectiveContainer:ArrangeChildrenVert(0)
	end
	
	nResult = math.max(nResult, self.knMinHeightEventItem)
	
	local nLeft, nTop, nRight, nBottom = wndQuest:GetAnchorOffsets()
	wndQuest:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nResult)
	return nResult
end

function LocalQuestTrackerAdvanced:OnResizeQuestObjective(wndObj)
	local nObjTextHeight = self.knInitialQuestObjectiveHeight
	
	-- If there's the spell icon is bigger, use that instead
	if wndObj:FindChild("SpellItemObjectiveBtn") or wndObj:GetName() == "SpellItem" then
		nObjTextHeight = math.max(nObjTextHeight, self.knInitialSpellItemHeight)
	end
	
	local wndQuestObjectiveText = wndObj:FindChild("QuestObjectiveText")
	
	-- If the text is bigger, use that instead
	if wndQuestObjectiveText then
		local nLocalWidth, nLocalHeight = wndQuestObjectiveText:SetHeightToContentHeight()
		nObjTextHeight = math.max(nObjTextHeight, nLocalHeight + 4) -- for lower g height

		-- Fake V-Align to match the button if it's just one line of text
		if wndObj:FindChild("SpellItemObjectiveBtn") and nLocalHeight < 20 then
			local nLeft, nTop, nRight, nBottom = wndQuestObjectiveText:GetAnchorOffsets()
			wndQuestObjectiveText:SetAnchorOffsets(nLeft, 9, nRight, nBottom)
		end
	end
	
	-- Also add extra height for Progress Bars
	if wndObj:FindChild("QuestProgressItem") then
		nObjTextHeight = nObjTextHeight + wndObj:FindChild("QuestProgressItem"):GetHeight()
	elseif wndObj:FindChild("PublicProgressItem") then
		nObjTextHeight = nObjTextHeight + wndObj:FindChild("PublicProgressItem"):GetHeight()
	end
	
	local nLeft, nTop, nRight, nBottom = wndObj:GetAnchorOffsets()
	wndObj:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nObjTextHeight)
	return nObjTextHeight
end

-----------------------------------------------------------------------------------------------
-- UI Interaction
-----------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:OnQuestCloseBtn(wndHandler, wndControl) -- wndHandler is "QuestCloseBtn" and its data is { wndQuest, tQuest }
	local queQuest = wndHandler:GetData().tQuest
	queQuest:SetActiveQuest(false)

	if queQuest:GetState() == Quest.QuestState_Botched then
		queQuest:Abandon()
	else
		queQuest:ToggleTracked()
	end
	
	self:QueueQuestForDestroy(queQuest)
	self:RedrawAll()
end

function LocalQuestTrackerAdvanced:OnMinimizedBtnChecked(wndHandler, wndControl, eMouseButton)
	if wndHandler:GetData() then
		self.tMinimized.tQuest[wndHandler:GetData()] = true
		self.tMinimized.tQuestTime[wndHandler:GetData()] = os.time()
	end

	self:RedrawAll()
end

function LocalQuestTrackerAdvanced:OnMinimizedBtnUnChecked(wndHandler, wndControl, eMouseButton)
	if wndHandler:GetData() then
		self.tMinimized.tQuest[wndHandler:GetData()] = nil
		self.tMinimized.tQuestTime[wndHandler:GetData()] = os.time()
	end

	self:RedrawAll()
end

function LocalQuestTrackerAdvanced:OnEpisodeMinimizedBtnChecked(wndHandler, wndControl, eMouseButton)
	if wndHandler:GetData() then
		self.tMinimized.tEpisode[wndHandler:GetData()] = true
		self.tMinimized.tEpisodeTime[wndHandler:GetData()] = os.time()
	end

	self:RedrawAll()
end

function LocalQuestTrackerAdvanced:OnEpisodeMinimizedBtnUnChecked(wndHandler, wndControl, eMouseButton)
	if wndHandler:GetData() then
		self.tMinimized.tEpisode[wndHandler:GetData()] = nil
		self.tMinimized.tEpisodeTime[wndHandler:GetData()] = os.time()
	end

	self:RedrawAll()
end

function LocalQuestTrackerAdvanced:OnQuestOpenLogBtn(wndHandler, wndControl) -- wndHandler should be "QuestOpenLogBtn" and its data is tQuest
	Event_FireGenericEvent("ShowQuestLog", wndHandler:GetData()) -- Codex (todo: deprecate this)
	Event_FireGenericEvent("GenericEvent_ShowQuestLog", wndHandler:GetData()) -- QuestLog
end

function LocalQuestTrackerAdvanced:OnQuestCallbackBtn(wndHandler, wndControl) -- wndHandler is "QuestCallbackBtn" and its data is tQuest
	CommunicatorLib.CallContact(wndHandler:GetData())
end

function LocalQuestTrackerAdvanced:OnShowEventStatsBtn(wndHandler, wndControl) -- wndHandler is "ShowEventStatsBtn" and its data is tEvent
	local peEvent = wndHandler:GetData() -- GOTCHA: Event Object is set up differently than the tZombieEvent table
	if peEvent and peEvent:HasLiveStats() then
		local tLiveStats = peEvent:GetLiveStats()
		Event_FireGenericEvent("GenericEvent_OpenEventStats", peEvent, peEvent:GetMyStats(), tLiveStats.arTeamStats, tLiveStats.arParticipantStats)
	end
end

function LocalQuestTrackerAdvanced:OnEventCallbackBtn(wndHandler, wndControl) -- wndHandler is "QuestCallbackBtn" and its data is wndEvent
	local wndEvent = wndHandler:GetData()
	for idx, tZombieEvent in pairs(self.tZombiePublicEvents) do
		if tZombieEvent.peEvent and tZombieEvent.peEvent == wndEvent:GetData() then
			if tZombieEvent.peEvent:GetEventType() == PublicEvent.PublicEventType_WorldEvent then
				Event_FireGenericEvent("GenericEvent_OpenEventStatsZombie", tZombieEvent)
			end

			table.remove(self.tZombiePublicEvents, idx)
			self:DestroyAndRedraw()
			return
		end
	end
end

function LocalQuestTrackerAdvanced:OnQuestHintArrow(wndHandler, wndControl, eMouseButton) -- wndHandler is "ControlBackerBtn" (can be from EventItem) and its data is wndQuest
	local wndQuest = wndHandler:GetData()

	if not wndQuest:FindChild("MinimizeBtn"):ContainsMouse() and (not wndQuest:FindChild("QuestCloseBtn") or not wndQuest:FindChild("QuestCloseBtn"):ContainsMouse()) then
		if eMouseButton == GameLib.CodeEnumInputMouse.Right and Apollo.IsShiftKeyDown() then
			Event_FireGenericEvent("GenericEvent_QuestLink", wndQuest:GetData())
		else
			wndQuest:GetData():ShowHintArrow()

			if self.tClickBlinkingQuest then
				self.tQuestTrackerBlink:Stop()
				self.tClickBlinkingQuest:SetActiveQuest(false)
			elseif self.tHoverBlinkingQuest then
				self.tHoverBlinkingQuest:SetActiveQuest(false)
			end

			if Quest.is(wndQuest:GetData()) then
				self.tClickBlinkingQuest = wndQuest:GetData()
				self.tClickBlinkingQuest:ToggleActiveQuest()
				self.tQuestTrackerBlink:Start()
			end
		end
	end
end

function LocalQuestTrackerAdvanced:OnQuestObjectiveHintArrow(wndHandler, wndControl, eMouseButton) -- "QuestObjectiveBtn" (can be from EventItem), data is { tQuest, tObjective.index }
	local tData = wndHandler:GetData()
	if tData and tData.peoObjective then
		tData.peoObjective:ShowHintArrow() -- Objectives do NOT default to parent if it fails
	elseif tData and tData.queOwner then
		tData.queOwner:ShowHintArrow(tData.nObjectiveIdx)

		if self.tClickBlinkingQuest then
			self.tQuestTrackerBlink:Stop()
			self.tClickBlinkingQuest:SetActiveQuest(false)
		elseif self.tHoverBlinkingQuest then
			self.tHoverBlinkingQuest:SetActiveQuest(false)
		end

		if Quest.is(tData.queOwner) then
			self.tClickBlinkingQuest = tData.queOwner
			self.tClickBlinkingQuest:ToggleActiveQuest()
			self.tQuestTrackerBlink:Start()
		end
	end

	return true -- Stop Propagation so the Quest Hint Arrow won't eat this call
end

-----------------------------------------------------------------------------------------------
-- Mouse Enter/Exits
-----------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:OnQuestItemMouseEnter(wndHandler, wndControl)
	if wndHandler == wndControl and wndHandler:GetData() and Quest.is(wndHandler:GetData()) then
		self.tHoverBlinkingQuest = wndHandler:GetData()

		if self.tClickBlinkingQuest == nil then
			self.tHoverBlinkingQuest:ToggleActiveQuest()
		end
	end
end

function LocalQuestTrackerAdvanced:OnQuestItemMouseExit(wndHandler, wndControl)
	if wndHandler == wndControl and wndHandler:GetData() and Quest.is(wndHandler:GetData()) then
		if self.tClickBlinkingQuest == nil and self.tHoverBlinkingQuest then
			self.tHoverBlinkingQuest:SetActiveQuest(false)
		end

		self.tHoverBlinkingQuest = nil
	end
end

function LocalQuestTrackerAdvanced:OnQuestNumberBackerMouseEnter(wndHandler, wndControl)
	if wndHandler == wndControl then
		wndHandler:FindChild("QuestNumberBackerGlow"):Show(true)
	end
end

function LocalQuestTrackerAdvanced:OnQuestNumberBackerMouseExit(wndHandler, wndControl)
	if wndHandler == wndControl then
		wndHandler:FindChild("QuestNumberBackerGlow"):Show(false)
	end
end

function LocalQuestTrackerAdvanced:OnControlBackerMouseEnter(wndHandler, wndControl) -- "ControlBackerBtn" of Quest or Event
	if wndHandler == wndControl then
		local wndQuest = wndHandler:GetData()
		local queQuest = wndQuest and wndQuest:GetData() or nil
		wndHandler:FindChild("MinimizeBtn"):Show(not queQuest or queQuest:GetState() ~= Quest.QuestState_Botched )
		if wndHandler:FindChild("QuestCloseBtn") then
			wndHandler:FindChild("QuestCloseBtn"):Show(true)
		end
	end
end

function LocalQuestTrackerAdvanced:OnControlBackerMouseExit(wndHandler, wndControl)
	if wndHandler == wndControl then
		wndHandler:FindChild("MinimizeBtn"):Show(false)
		if wndHandler:FindChild("QuestCloseBtn") then
			wndHandler:FindChild("QuestCloseBtn"):Show(false)
		end
	end
end

function LocalQuestTrackerAdvanced:OnEpisodeControlBackerMouseEnter(wndHandler, wndControl)
	if wndHandler == wndControl then
		wndHandler:FindChild("EpisodeMinimizeBtn"):Show(true)
	end
end

function LocalQuestTrackerAdvanced:OnEpisodeControlBackerMouseExit(wndHandler, wndControl)
	if wndHandler == wndControl then
		wndHandler:FindChild("EpisodeMinimizeBtn"):Show(false)
	end
end

-----------------------------------------------------------------------------------------------
-- Code Events (mostly removing zombies)
-----------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:OnShowResurrectDialog()
	unitPlayer = GameLib.GetPlayerUnit()
	if unitPlayer then
		self.bPlayerIsDead = unitPlayer:IsDead()
	end
end

function LocalQuestTrackerAdvanced:OnPlayerResurrected()
	self.bPlayerIsDead = false
end

function LocalQuestTrackerAdvanced:OnToggleLongQuestText(bToggle)
	self.bShowLongQuestText = bToggle
end

function LocalQuestTrackerAdvanced:OnQuestStateChanged(queQuest, eState)
	if not self.wndMain then
		return
	end

	if eState == Quest.QuestState_Completed or eState == Quest.QuestState_Abandoned or eState == Quest.QuestState_Botched or eState == Quest.QuestState_Unknown then
		self:QueueQuestForDestroy(queQuest)
	else
		self.nFlashThisQuest = queQuest
	end

	self:RequestRedrawAll()
end

function LocalQuestTrackerAdvanced:OnQuestObjectiveUpdated(queQuest, nObjective)
	if not queQuest or not queQuest:IsTracked() or queQuest:ObjectiveIsVisible(nObjective) == false then
		return
	end
	
	self.tActiveProgBarQuests[queQuest:GetId()] = os.clock()
	Apollo.CreateTimer("QuestTracker_EarliestProgBarTimer", knQuestProgBarFadeoutTime, false)
	-- GOTCHA: Apollo quirk, if you don't StopTimer before this, only the earliest is caught. So check and refire event in the handler.

	self:OnDestroyQuestObject(queQuest)
end

function LocalQuestTrackerAdvanced:OnQuestTracker_EarliestProgBarTimer()
	-- GOTCHA: Apollo quirk, only the earliest is caught. So check and refire event if applicable.
	local nComparisonTime = os.clock()
	local nLowestTime = 9000
	for nCurrQuestId, nCurrTime in pairs(self.tActiveProgBarQuests) do
		if (nCurrTime + knQuestProgBarFadeoutTime) < (nComparisonTime + 1) then -- Plus one for safety
			self.tActiveProgBarQuests[nCurrQuestId] = nil
		else
			local nDifference = (nCurrTime + knQuestProgBarFadeoutTime) - nComparisonTime
			nLowestTime = nDifference < nLowestTime and nDifference or nLowestTime
		end
	end

	if nLowestTime ~= 9000 then
		Apollo.CreateTimer("QuestTracker_EarliestProgBarTimer", nLowestTime, false)
	end
end

function LocalQuestTrackerAdvanced:OnDestroyQuestObject(queQuest)
	self.nFlashThisQuest = queQuest
	self:QueueQuestForDestroy(queQuest)
	self:RedrawAll()
end

function LocalQuestTrackerAdvanced:OnGenericEvent_ChallengeTrackerToggled(bVisible)
	if not self.wndMain or self.bChallengeVisible == bVisible then
		return
	end

	self.bChallengeVisible = bVisible
	
	if self.settings.bHideWhileChallenge then
		-- Hides Quest Tracker while Challenge Tracker is visible
		self.wndMain:Show(not self.bChallengeVisible)
	end
	
	self:SetPosition()
end

-----------------------------------------------------------------------------------------------
-- Public Events
-----------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:OnPublicEventStart(peEvent)
	-- Remove from zombie list if we're restarting it
	for idx, tZombieEvent in pairs(self.tZombiePublicEvents) do
		if tZombieEvent.peEvent == peEvent then
			self.tZombiePublicEvents[idx] = nil
			local wndPublicEvent = self.wndQuestTrackerScroll:FindChildByUserData(kstrPublicEventMarker)
			if wndPublicEvent then
				local wndEvent = wndPublicEvent:FindChildByUserData(peEvent)
				if wndEvent then
					wndEvent:Destroy()
					self:OnEventDestroyed(peEvent)
				end
			end
			break
		end
	end
	
	self.tDestroyAndRedraw:Start()
end

function LocalQuestTrackerAdvanced:OnPublicEventEnd(peEvent, eReason, tStats)
	-- Add to list, or delete if we left the area
	if (eReason == PublicEvent.PublicEventParticipantRemoveReason_CompleteSuccess or eReason == PublicEvent.PublicEventParticipantRemoveReason_CompleteFailure)
	and peEvent:GetEventType() ~= PublicEvent.PublicEventType_SubEvent then
		table.insert(self.tZombiePublicEvents, {["peEvent"] = peEvent, ["eReason"] = eReason, ["tStats"] = tStats, ["nTimer"] = self.ZombieTimerMax})
	end

	-- Delete existing
	local wndPublicEvent = self.wndQuestTrackerScroll:FindChildByUserData(kstrPublicEventMarker)
	if wndPublicEvent then
		local wndEvent = wndPublicEvent:FindChildByUserData(peEvent)
		if wndEvent then
			wndEvent:Destroy()
			self:OnEventDestroyed(peEvent)
		end
	end
	self:RequestRedrawAll()
end

function LocalQuestTrackerAdvanced:OnPublicEventUpdate(tObjective)
	if tObjective:GetEvent() then
		self.tPublicEventsToRedraw[tObjective:GetEvent():GetName()] = true
		self:RequestRedrawAll()
	end
end

-----------------------------------------------------------------------------------------------
-- String Building
-----------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:HelperShowQuestCallbackBtn(wndQuest, queQuest, strNumberBackerArt, strCallbackBtnArt)
	wndQuest:FindChild("QuestNumberBackerArt"):SetSprite(strNumberBackerArt)
	
	local tContactInfo = queQuest:GetContactInfo()
	
	if not queQuest:IsCommunicatorReceived() or queQuest:IsCommunicatorReceivedFromRec() then
		if not tContactInfo then
			self.tQueuedCommMessages[queQuest:GetId()] = {wndQuest = wndQuest, queQuest = queQuest, strNumberBackerArt = strNumberBackerArt, strCallbackBtnArt = strCallbackBtnArt}
			return
		else
			self.tQueuedCommMessages[queQuest:GetId()] = nil
		end
	end
	
	if not tContactInfo or not tContactInfo.strName or string.len(tContactInfo.strName) <= 0 then
		return
	end

	local strName = String_GetWeaselString(Apollo.GetString("QuestTracker_ContactName"), tContactInfo.strName)
	wndQuest:FindChild("QuestCompletedBacker"):Show(true)
	wndQuest:FindChild("QuestCallbackBtn"):ChangeArt(strCallbackBtnArt)
	wndQuest:FindChild("QuestCallbackBtn"):Enable(not self.bPlayerIsDead)
	wndQuest:FindChild("QuestCallbackBtn"):SetTooltip(string.format("<P Font=\"CRB_InterfaceMedium\">%s</P>", strName))
end

function LocalQuestTrackerAdvanced:HelperBuildTimedQuestTitle(queQuest)
	local strTitle = queQuest:GetTitle()
	strTitle = string.format("<T Font=\"CRB_InterfaceMedium\" TextColor=\"%s\">%s</T>", kstrLightGrey, strTitle)
	strTitle = self:HelperPrefixTimeString(math.max(0, math.floor(queQuest:GetQuestTimeRemaining() / 1000)), strTitle)
	
	return strTitle
end

function LocalQuestTrackerAdvanced:BuildObjectiveTitleString(queQuest, tObjective, bIsTooltip)
	local strResult = ""

	-- Early exit for completed
	if queQuest:GetState() == Quest.QuestState_Achieved then
		strResult = queQuest:GetCompletionObjectiveShortText()
		if bIsTooltip or self.bShowLongQuestText or not strResult or string.len(strResult) <= 0 then
			strResult = queQuest:GetCompletionObjectiveText()
		end
		return string.format("<T Font=\"CRB_InterfaceMedium\">%s</T>", strResult)
	end

	-- Use short form or reward text if possible
	local strShortText = queQuest:GetObjectiveShortDescription(tObjective.nIndex)
	if self.bShowLongQuestText or bIsTooltip then
		strResult = string.format("<T Font=\"CRB_InterfaceMedium\">%s</T>", tObjective.strDescription)
	elseif strShortText and string.len(strShortText) > 0 then
		strResult = string.format("<T Font=\"CRB_InterfaceMedium\">%s</T>", strShortText)
	end

	-- Prefix Optional or Progress if it hasn't been finished yet
	if tObjective.nCompleted < tObjective.nNeeded then
		local strPrefix = ""
		if tObjective and not tObjective.bIsRequired then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium\">%s</T>", Apollo.GetString("QuestLog_Optional"))
			strResult = String_GetWeaselString(Apollo.GetString("QuestTracker_BuildText"), strPrefix, strResult)
		end

		-- Use Percent if Progress Bar
		if tObjective.nNeeded > 1 and queQuest:DisplayObjectiveProgressBar(tObjective.nIndex) then
			local strColor = self.tActiveProgBarQuests[queQuest:GetId()] and kstrHighlight or "ffffffff"
			local strPercentComplete = String_GetWeaselString(Apollo.GetString("QuestTracker_PercentComplete"), tObjective.nCompleted)
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\" TextColor=\"%s\">%s</T>", strColor, strPercentComplete)
			strResult = String_GetWeaselString(Apollo.GetString("QuestTracker_BuildText"), strPrefix, strResult)
		elseif tObjective.nNeeded > 1 then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\">%s</T>", String_GetWeaselString(Apollo.GetString("QuestTracker_ValueComplete"), tObjective.nCompleted, tObjective.nNeeded))
			strResult = String_GetWeaselString(Apollo.GetString("QuestTracker_BuildText"), strPrefix, strResult)
		end
	end

	-- Prefix time for timed objectives
	if queQuest:IsObjectiveTimed(tObjective.nIndex) then
		strResult = self:HelperPrefixTimeString(math.max(0, math.floor(queQuest:GetObjectiveTimeRemaining(tObjective.nIndex) / 1000)), strResult)
	end

	return strResult
end

function LocalQuestTrackerAdvanced:BuildEventObjectiveTitleString(queQuest, peoObjective, bIsTooltip)
	-- Use short form or reward text if possible
	local strResult = ""
	local strShortText = peoObjective:GetShortDescription()
	if strShortText and string.len(strShortText) > 0 and not bIsTooltip then
		strResult = string.format("<T Font=\"CRB_InterfaceMedium\">%s</T>", strShortText)
	else
		strResult = string.format("<T Font=\"CRB_InterfaceMedium\">%s</T>", peoObjective:GetDescription())
	end

	-- Progress Brackets and Time if Active
	if peoObjective:GetStatus() == PublicEventObjective.PublicEventStatus_Active then
		local nCompleted = peoObjective:GetCount()
		local eCategory = peoObjective:GetCategory()
		local eType = peoObjective:GetObjectiveType()
		local nNeeded = peoObjective:GetRequiredCount()

		-- Prefix Brackets
		local strPrefix = ""
		if nNeeded == 0 and (eType == PublicEventObjective.PublicEventObjectiveType_Exterminate or eType == PublicEventObjective.PublicEventObjectiveType_DefendObjectiveUnits) then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\">%s </T>", String_GetWeaselString(Apollo.GetString("QuestTracker_Remaining"), nCompleted))
		elseif eType == PublicEventObjective.PublicEventObjectiveType_DefendObjectiveUnits and not peoObjective:ShowPercent() and not peoObjective:ShowHealthBar() then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\">%s </T>", String_GetWeaselString(Apollo.GetString("QuestTracker_Remaining"), (nCompleted - nNeeded + 1)))
		elseif eType == PublicEventObjective.PublicEventObjectiveType_Turnstile then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\">%s </T>", String_GetWeaselString(Apollo.GetString("QuestTracker_WaitingForMore"), math.abs(nCompleted - nNeeded)))
		elseif eType == PublicEventObjective.PublicEventObjectiveType_ParticipantsInTriggerVolume then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\">%s </T>",  String_GetWeaselString(Apollo.GetString("QuestTracker_WaitingForMore"), math.abs(nCompleted - nNeeded)))
		elseif eType == PublicEventObjective.PublicEventObjectiveType_TimedWin then
			-- Do Nothing
		elseif nNeeded > 1 and not peoObjective:ShowPercent() and not peoObjective:ShowHealthBar() then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\">%s </T>", String_GetWeaselString(Apollo.GetString("QuestTracker_ValueComplete"), nCompleted, nNeeded))
		end

		if strPrefix ~= "" then
			strResult = String_GetWeaselString(Apollo.GetString("QuestTracker_BuildText"), strPrefix, strResult)
			strPrefix = ""
		end

		-- Prefix Time
		if peoObjective:IsBusy() then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\" TextColor=\"%s\">%s </T>", kstrYellow, Apollo.GetString("QuestTracker_Paused"))
			strResult = String_GetWeaselString(Apollo.GetString("QuestTracker_BuildText"), strPrefix, strResult)
			strPrefix = ""
		elseif peoObjective:GetTotalTime() > 0 then
			local strColorOverride = nil
			if peoObjective:GetObjectiveType() == PublicEventObjective.PublicEventObjectiveType_TimedWin then
				strColorOverride = kstrGreen
			end
			strResult = self:HelperPrefixTimeString(math.max(0, math.floor((peoObjective:GetTotalTime() - peoObjective:GetElapsedTime()) / 1000)), strResult, strColorOverride)
		end

		-- Extra formatting
		if eCategory == PublicEventObjective.PublicEventObjectiveCategory_PlayerPath then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\">%s </T>", String_GetWeaselString(Apollo.GetString("CRB_ProgressSimple"), self.strPlayerPath or Apollo.GetString("CRB_Path")))
		elseif eCategory == PublicEventObjective.PublicEventObjectiveCategory_Optional then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\">%s </T>", Apollo.GetString("QuestTracker_OptionalTag"))
		elseif eCategory == PublicEventObjective.PublicEventObjectiveCategory_Challenge then
			strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\">%s </T>", Apollo.GetString("QuestTracker_ChallengeTag"))
		end

		if strPrefix ~= "" then
			strResult = String_GetWeaselString(Apollo.GetString("QuestTracker_BuildText"), strPrefix, strResult)
		end
	end
	return strResult
end

-----------------------------------------------------------------------------------------------
-- PvP
-----------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:OnLeavePvP()
	self.bDrawPvPScreenOnly = false
	if self.wndMain:FindChild("SwapToPvP") and self.wndMain:FindChild("SwapToPvP"):IsValid() then
		self.wndMain:FindChild("SwapToPvP"):Destroy()
	end
	if self.wndMain:FindChild("SwapToQuests") and self.wndMain:FindChild("SwapToQuests"):IsValid() then
		self.wndMain:FindChild("SwapToQuests"):Destroy()
	end
	self:DestroyAndRedraw()
end

function LocalQuestTrackerAdvanced:OnSwapToPvPBtn() -- Also from code
	self.bDrawPvPScreenOnly = true
	if self.wndMain:FindChild("SwapToPvP") and self.wndMain:FindChild("SwapToPvP"):IsValid() then
		self.wndMain:FindChild("SwapToPvP"):Destroy()
	end
	self:FactoryProduce(self.wndQuestTrackerScroll, "SwapToQuests", "SwapToQuests")
	self:DestroyAndRedraw()
end

function LocalQuestTrackerAdvanced:OnSwapToDungeonsBtn() -- Also from code
	self.bDrawDungeonScreenOnly = true
	if self.wndMain:FindChild("SwapToDungeons") and self.wndMain:FindChild("SwapToDungeons"):IsValid() then
		self.wndMain:FindChild("SwapToDungeons"):Destroy()
	end
	self:FactoryProduce(self.wndQuestTrackerScroll, "SwapToQuests", "SwapToQuests")
	self:DestroyAndRedraw()
end

function LocalQuestTrackerAdvanced:OnSwapToQuestsBtn()
	if self.bDrawPvPScreenOnly then
		self.bDrawPvPScreenOnly = false
		self:FactoryProduce(self.wndQuestTrackerScroll, "SwapToPvP", "SwapToPvP")
	end

	if self.bDrawDungeonScreenOnly then -- TODO investigate what happens when both are active
		self.bDrawDungeonScreenOnly = false
		self:FactoryProduce(self.wndQuestTrackerScroll, "SwapToDungeons", "SwapToDungeons")
	end

	if self.wndMain:FindChild("SwapToQuests") and self.wndMain:FindChild("SwapToQuests"):IsValid() then
		self.wndMain:FindChild("SwapToQuests"):Destroy()
	end
	self:RedrawAll() -- GOTCHA: Don't destroy, we check for SwapToPvPBtn being valid later
end

function LocalQuestTrackerAdvanced:OnGenerateTooltip(wndControl, wndHandler, eType, arg1, arg2)
	local xml = nil
	if eType == Tooltip.TooltipGenerateType_ItemInstance then -- Doesn't need to compare to item equipped
		if Tooltip ~= nil and Tooltip.GetItemTooltipForm~= nil then
			Tooltip.GetItemTooltipForm(self, wndControl, arg1, {})
		end
	elseif eType == Tooltip.TooltipGenerateType_ItemData then -- Doesn't need to compare to item equipped
		if Tooltip ~= nil and Tooltip.GetItemTooltipForm~= nil then
			Tooltip.GetItemTooltipForm(self, wndControl, arg1, {})
		end
	elseif eType == Tooltip.TooltipGenerateType_GameCommand then
		xml = XmlDoc.new()
		xml:AddLine(arg2)
		wndControl:SetTooltipDoc(xml)
	elseif eType == Tooltip.TooltipGenerateType_Macro then
		xml = XmlDoc.new()
		xml:AddLine(arg1)
		wndControl:SetTooltipDoc(xml)
	elseif eType == Tooltip.TooltipGenerateType_Spell then
		if Tooltip ~= nil and Tooltip.GetSpellTooltipForm ~= nil then
			Tooltip.GetSpellTooltipForm(self, wndControl, arg1)
		end
	elseif eType == Tooltip.TooltipGenerateType_PetCommand then
		xml = XmlDoc.new()
		xml:AddLine(arg2)
		wndControl:SetTooltipDoc(xml)
	end
end

-----------------------------------------------------------------------------------------------
-- Helpers
-----------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:HelperPrefixTimeString(fTime, strAppend, strColorOverride)
	local fSeconds = fTime % 60
	local fMinutes = fTime / 60
	local strColor = kstrYellow
	if strColorOverride then
		strColor = strColorOverride
	elseif fMinutes < 1 and fSeconds <= 30 then
		strColor = kstrRed
	end
	local strPrefix = string.format("<T Font=\"CRB_InterfaceMedium_B\" TextColor=\"%s\">(%d:%.02d)</T>", strColor, fMinutes, fSeconds)
	return String_GetWeaselString(Apollo.GetString("QuestTracker_BuildText"), strPrefix, strAppend)
end

function LocalQuestTrackerAdvanced:HelperFindAndDestroyQuests()
	if not #self.tQuestsQueuedForDestroy then
		return
	end
		
	for idx1, wndEpGroup in pairs(self.wndQuestTrackerScroll:GetChildren()) do
		if wndEpGroup:GetName() == "EpisodeGroupItem" then
			for idx2, wndEp in pairs(wndEpGroup:FindChild("EpisodeGroupContainer"):GetChildren()) do
				for key, queQuest in pairs(self.tQuestsQueuedForDestroy) do
					if wndEp:GetName() == "EpisodeItem" then
						local wndQuest = wndEp:FindChild("EpisodeQuestContainer"):FindChildByUserData(queQuest)
						if wndQuest then
							wndQuest:Destroy()
	
							if wndEp:GetData() ~= kstrPublicEventMarker and next(wndEp:GetData():GetTrackedQuests()) == nil then
								wndEp:Destroy()
							end
	
							if next(wndEpGroup:FindChild("EpisodeGroupContainer"):GetChildren()) == nil then
								wndEpGroup:Destroy()
							end
	
							table.remove(self.tQuestsQueuedForDestroy, key)
							if #self.tQuestsQueuedForDestroy == 0 then
								return
							end
							
							self:OnQuestDestroyed(queQuest)
						end
					elseif wndEp:GetName() == "QuestItem" and wndEp:GetData() == queQuest then
						wndEp:Destroy()
	
						if next(wndEpGroup:FindChild("EpisodeGroupContainer"):GetChildren()) == nil then
							wndEpGroup:Destroy()
						end
	
						table.remove(self.tQuestsQueuedForDestroy, key)
						if #self.tQuestsQueuedForDestroy == 0 then
							return
						end
						
						self:OnQuestDestroyed(queQuest)
					end
				end
			end
		end
	end
	
	-- anything left in this list doesn't exist anymore
	self.tQuestsQueuedForDestroy = {}
end

function LocalQuestTrackerAdvanced:OnQuestDestroyed(queQuest)
	for index, tQuestInfo in pairs(self.tTimedQuests) do
		if tQuestInfo.queQuest == queQuest then
			table.remove(self.tTimedQuests, index)
		end
	end
	
	for index, tObjectiveInfo in pairs(self.tTimedObjectives) do
		if tObjectiveInfo.queQuest == queQuest then
			table.remove(self.tTimedObjectives, index)
		end
	end
	
	self.tQueuedCommMessages[queQuest:GetId()] = nil
end

function LocalQuestTrackerAdvanced:OnEventDestroyed(peEvent)
	for index, tEventInfo in pairs(self.tTimedEvents) do
		if tEventInfo.peEvent == peEvent then
			table.remove(self.tTimedEvents, index)
		end
	end
	
	for index, tEventObjectiveInfo in pairs(self.tTimedEventObjectives) do
		if tEventObjectiveInfo.peEvent == peEvent then
			table.remove(self.tTimedEventObjectives, index)
		end
	end
end

function LocalQuestTrackerAdvanced:FactoryProduce(wndParent, strFormName, tObject)
	local wnd = wndParent:FindChildByUserData(tObject)
	if not wnd then
		wnd = Apollo.LoadForm(self.xmlDoc, strFormName, wndParent, self)
		wnd:SetData(tObject)
	end
	return wnd
end

---------------------------------------------------------------------------------------------------
-- Tutorial anchor request
---------------------------------------------------------------------------------------------------
function LocalQuestTrackerAdvanced:OnTutorial_RequestUIAnchor(eAnchor, idTutorial, strPopupText)
	if eAnchor == GameLib.CodeEnumTutorialAnchor.QuestTracker or eAnchor == GameLib.CodeEnumTutorialAnchor.QuestCommunicatorReceived then

	local tRect = {}
	tRect.l, tRect.t, tRect.r, tRect.b = self.wndMain:GetRect()

	Event_FireGenericEvent("Tutorial_RequestUIAnchorResponse", eAnchor, idTutorial, strPopupText, tRect)
	end
end

---------------------------------------------------------------------------------------------------
-- LocalQuestTrackerAdvanced - New Functions
---------------------------------------------------------------------------------------------------
function LocalQuestTrackerAdvanced:UpdateTimeouts(timeList, booleanList)
	-- Never timeout
	if self.settings.nTimeout <= 0 then
		return false
	end
	
	local timeout = false
	for idx, x in pairs(timeList) do
		if (x ~= nil and (os.time() - x) > self.settings.nTimeout) then
			log:debug(string.format("Timed out: %d", idx))
			timeList[idx] = nil
			booleanList[idx] = nil
			timeout = true
		end
	end
	
	return timeout
end

function LocalQuestTrackerAdvanced:ResetHide(timeList, booleanList)
	local timeout = false
	for idx, x in pairs(booleanList) do
		timeList[idx] = nil
		booleanList[idx] = nil
	end
	
	return timeout
end

function LocalQuestTrackerAdvanced:OnChangeZoneName(oVar, strNewZone)
	self.currentZone = GameLib.GetCurrentZoneMap().strName
	self.currentZoneId = GameLib.GetCurrentZoneMap().id
	log:debug(string.format("Changed zone to %s (%d)",self.currentZone, self.currentZoneId))
	
	self:AutoTrack()
end

function LocalQuestTrackerAdvanced:OnChangeSubZoneName(oVar, strNewZone)
	self.currentZone = GameLib.GetCurrentZoneMap().strName
	self.currentZoneId = GameLib.GetCurrentZoneMap().id
	self.currentSubZone = strNewZone
	log:debug(string.format("Changed subzone to %s (unknown)",self.currentSubZone))

	self:AutoTrack()
end

function LocalQuestTrackerAdvanced:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	
	log:debug("Saving Settings...")
    log:debug(self.settings)
	
	self.settings.bFirstRun = false
	return self.settings
end

function LocalQuestTrackerAdvanced:OnRestore(eType, tData)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	
    log:debug("Loading Settings...")
    log:debug(tData)

	if tData then
		self.settings = VILib:MergeTables(self.settings, tData)
	end
	if self.wndmain then
		self:ApplyFreePositionSize()
		self:ApplyFilterPreferences()
	end
end


---------------------------------------------------------------------------------------------------
-- Settings Functions
---------------------------------------------------------------------------------------------------

---------------------------------------------------------------------------------------------------
-- Basic Functions
---------------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:OnSlashCommand(strCmd, strParam)
	if strParam ~= "" then
		if string.lower(strParam) == "reset" then	
			for k, v in pairs(self.settings.tWindowPositions.tMainOffsets) do
				self.settings.tWindowPositions.tMainOffsets[k] = {0, 0}
			end
			self.settings.tWindowPositions.nMainAnchor = 7
			self.settings.tWindowPositions.tMainOffsets[7] = {15, -95}
			if self.bSettingsOpen then
				self:SetSettings()
			end
			self:SetPosition()
			VILib:WriteToChat("all window positions have been reset.")
		elseif string.lower(strParam) == "autohide 1" then
			self.settings.bHideWhileChallenge = true
			self.wndMain:Show(not self.bChallengeVisible)
		elseif string.lower(strParam) == "autohide 0" then
			self.settings.bHideWhileChallenge = false
			self.wndMain:Show(true)
		elseif string.lower(strParam) == "log" then
			log:SetLevel("DEBUG")
			if self.currentZone == nil then
				self.currentZone = GameLib.GetCurrentZoneMap().strName
			end
			log:debug("Current Zone: "..self.currentZone)
			local tCategoryEpisodeHaveQuestsCache = {}
			local tCategoryHaveQuestsCache = {}
			for idx1, qcCategory in pairs(QuestLib.GetKnownCategories()) do
				if self:DoesCategoryHaveQuests(qcCategory, tCategoryHaveQuestsCache) then
					for idx2, epiEpisode in pairs(qcCategory:GetEpisodes()) do
						if self:DoesCategoryEpisodeHaveQuests(qcCategory, epiEpisode, tCategoryEpisodeHaveQuestsCache) and (epiEpisode:GetZoneName() == "" or epiEpisode:GetZoneName() == self.currentZone) then
							for idx3, queQuest in pairs(epiEpisode:GetAllQuests(qcCategory:GetId())) do
								if (queQuest:GetState() ~= Quest.QuestState_Completed) and not queQuest:IsIgnored() and queQuest:GetCategory():GetId() ~= 53 then
									log:debug("Quest: "..queQuest:GetTitle()..", Category: "..queQuest:GetCategory():GetTitle())
								end
							end
						end
					end
				end
			end
			log:SetLevel("INFO")
		end
	else
		self:ShowHideSettings()
	end
end

function LocalQuestTrackerAdvanced:OnMouseDownCatcher( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation )
end

---------------------------------------------------------------------------------------------------
-- LocalQuestTrackerAdvancedForm Functions
---------------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:OnQuestTrackerMove( wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom )
	local nL, nT, nR, nB = self.wndMain:GetAnchorOffsets()
	local nWidth, nHeight = VILib:Round(nR - nL), VILib:Round(nB - nT)
	local fScale = 1 - (1 / Apollo.GetConsoleVariable("ui.Scale"))
	local nOffsetX, nOffsetY = VILib:Round(math.abs(nWidth * fScale) / 2), VILib:Round(math.abs(nHeight * fScale) / 2)
	local noL, noT, noR, noB = self.wndMain:GetAnchorPoints()
	self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][1] = nL + nWidth*noL + nOffsetX
	self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][2] = nT + nHeight*noT + nOffsetY
	self.settings.nWidth = nWidth
	self.settings.nHeight = nHeight
	if self.wndSettings then
		self.wndSettingsSizePosition:FindChild("Size:Width:Slider"):SetValue(self.settings.nWidth)
		self.wndSettingsSizePosition:FindChild("Size:Width:Value"):SetText(self.settings.nWidth)
		self.wndSettingsSizePosition:FindChild("Size:Height:Slider"):SetValue(self.settings.nHeight)
		self.wndSettingsSizePosition:FindChild("Size:Height:Value"):SetText(self.settings.nHeight)
		self.wndSettingsSizePosition:FindChild("Position:PosOffsetX:Slider"):SetValue(self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][1])
		self.wndSettingsSizePosition:FindChild("Position:PosOffsetX:Value"):SetText(self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][1])
		self.wndSettingsSizePosition:FindChild("Position:PosOffsetY:Slider"):SetValue(self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][2])
		self.wndSettingsSizePosition:FindChild("Position:PosOffsetY:Value"):SetText(self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][2])
	end
end



---------------------------------------------------------------------------------------------------
-- SettingsForm Functions
---------------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:OnConfigure()
	self:ShowHideSettings()
end

function LocalQuestTrackerAdvanced:OnCloseButton( wndHandler, wndControl, eMouseButton )
	self:ShowHideSettings()
end

function LocalQuestTrackerAdvanced:OnSettingsWindowClosed( wndHandler, wndControl )
	if wndHandler ~= wndControl then return end
	self:ShowHideSettings()
end

function LocalQuestTrackerAdvanced:ShowHideSettings()
	self.bSettingsOpen = not self.bSettingsOpen
	if not self.wndSettings then
		-- Create Window
		self.wndSettings = Apollo.LoadForm(self.xmlDocSettings, "SettingsForm", nil, self)
		self.wndSettingsSizePosition = Apollo.LoadForm(self.xmlDocSettings, "SettingsSizePosition", self.wndSettings:FindChild("Content"), self)
		self.wndSettingsFilterPreferences = Apollo.LoadForm(self.xmlDocSettings, "SettingsFilterPreferences", self.wndSettings:FindChild("Content"), self)
		self.wndSettings:FindChild("Menu:SizePositionButton"):SetCheck(true)
		self:OnSettingsMenuButtonCheck()
		self.wndSettings:FindChild("Menu:VersionText"):SetText("Version " .. version)
	end
	if self.bSettingsOpen then
		-- Open Window
		self:SetSettings()
		self.wndSettings:ToFront()
		local nScreenWidth, nScreenHeight = Apollo.GetScreenSize()
		local nLeft, nTop, nRight, nBottom = self.wndSettings:GetRect()
		local nWidth, nHeight = nRight - nLeft, nBottom - nTop
		if self.settings.tWindowPositions.tSettings[1] then
			self.wndSettings:Move((nScreenWidth * self.settings.tWindowPositions.tSettings[1]), (nScreenHeight * self.settings.tWindowPositions.tSettings[2]), nWidth, nHeight)
		else
			self.wndSettings:Move(((nScreenWidth / 2) - (nWidth / 2)), ((nScreenHeight / 2) - (nHeight / 2)), nWidth, nHeight)
		end
		SettingsWindowRef = self.wndSettings
		SettingsRef = self.settings
		QuestTrackerRef = self.wndQuestTracker
	else
		-- Close Window
	end
	self.wndSettings:Show(self.bSettingsOpen, false)
end

function LocalQuestTrackerAdvanced:SetSettings()
	self.wndSettingsSizePosition:FindChild("Size:FreeScaling:Button"):SetCheck(self.settings.bFreeScaling)
	self.wndSettingsSizePosition:FindChild("Size:Width:Slider"):SetValue(self.settings.nWidth)
	self.wndSettingsSizePosition:FindChild("Size:Width:Value"):SetText(self.settings.nWidth)
	self.wndSettingsSizePosition:FindChild("Size:Height:Slider"):SetValue(self.settings.nHeight)
	self.wndSettingsSizePosition:FindChild("Size:Height:Value"):SetText(self.settings.nHeight)
	
	self:FillAnchorList()
	self.wndSettingsSizePosition:FindChild("Position:FreePositioning:Button"):SetCheck(self.settings.bFreePositioning)
	self.wndSettingsSizePosition:FindChild("Position:PosOffsetX:Slider"):SetValue(self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][1])
	self.wndSettingsSizePosition:FindChild("Position:PosOffsetX:Value"):SetText(self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][1])
	self.wndSettingsSizePosition:FindChild("Position:PosOffsetY:Slider"):SetValue(self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][2])
	self.wndSettingsSizePosition:FindChild("Position:PosOffsetY:Value"):SetText(self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][2])

	self.wndSettingsSizePosition:FindChild("TextPosition:TopAlignment:Button"):SetCheck(self.settings.bTopAlign)


	self.wndSettingsFilterPreferences:FindChild("Collapse:Episodes:Button"):SetCheck(self.settings.bAutoCollapseEpisodes)
	self.wndSettingsFilterPreferences:FindChild("Collapse:Quests:Button"):SetCheck(self.settings.bAutoCollapseQuests)
	self.wndSettingsFilterPreferences:FindChild("Collapse:Timeout:Slider"):SetValue(self.settings.nTimeout)
	self.wndSettingsFilterPreferences:FindChild("Collapse:Timeout:Value"):SetText(self.settings.nTimeout)

	self.wndSettingsFilterPreferences:FindChild("Track:Quests:WorldStoryButton"):SetCheck(self.settings.bAutoTrackWorldStory)
	self.wndSettingsFilterPreferences:FindChild("Track:Quests:ZoneStoryButton"):SetCheck(self.settings.bAutoTrackZoneStory)
	self.wndSettingsFilterPreferences:FindChild("Track:Quests:RegionalStoryButton"):SetCheck(self.settings.bAutoTrackRegionalStory)
	self.wndSettingsFilterPreferences:FindChild("Track:Quests:TaskButton"):SetCheck(self.settings.bAutoTrackTasks)
	
	self.wndSettingsFilterPreferences:FindChild("ShowHide:Tradeskills:Button"):SetCheck(self.settings.bAutoShowHideTradeskills)
	self.wndSettingsFilterPreferences:FindChild("ShowHide:Episodes:Button"):SetCheck(self.settings.bAutoShowHideEpisodes)
end

function LocalQuestTrackerAdvanced:FillAnchorList()
	local nSel = 0
	self.wndSettingsSizePosition:FindChild("AnchorDropdownParent:AnchorGrid"):DeleteAll()
	for i, v in ipairs(tAnchors) do
		if self.settings.tWindowPositions.nMainAnchor == i then
			self.wndSettingsSizePosition:FindChild("Position:Anchor:DropdownButton"):SetText(v["sLabel"])
			nSel = i
		end
		self.wndSettingsSizePosition:FindChild("AnchorDropdownParent:AnchorGrid"):AddRow(v["sLabel"], "", v)
	end
	self.wndSettingsSizePosition:FindChild("AnchorDropdownParent:AnchorGrid"):SetCurrentRow(nSel)
end

function LocalQuestTrackerAdvanced:OnSettingsWindowHide( wndHandler, wndControl )
	if wndHandler ~= wndControl then return end
	if self.wndSettings then
		self.wndSettings:Destroy()
		self.wndSettings = nil
		SettingsWindowRef = nil
		SettingsRef = nil
		QuestTrackerRef = nil
	end
end

function LocalQuestTrackerAdvanced:OnSettingsWindowMove( wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom )
	local nScreenWidth, nScreenHeight = Apollo.GetScreenSize()
	local nPosLeft, nPosTop = self.wndSettings:GetPos()
	self.settings.tWindowPositions.tSettings[1], self.settings.tWindowPositions.tSettings[2] = nPosLeft / nScreenWidth, nPosTop / nScreenHeight
end

function LocalQuestTrackerAdvanced:OnSettingsMenuButtonCheck( wndHandler, wndControl, eMouseButton )
	self.wndSettingsSizePosition:Show(self.wndSettings:FindChild("Menu:SizePositionButton"):IsChecked())
	self.wndSettingsFilterPreferences:Show(self.wndSettings:FindChild("Menu:FilterPreferencesButton"):IsChecked())
	
	self.wndSettings:FindChild("Content"):SetVScrollPos(0)
	self.wndSettings:FindChild("Content"):RecalculateContentExtents()
end

---------------------------------------------------------------------------------------------------
-- SettingsSizePosition Functions
---------------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:ApplyFreePositionSize()
	self.wndMain:FindChild("Background"):SetSprite((self.settings.bFreePositioning or self.settings.bFreeScaling) and "BK3:UI_BK3_Holo_InsetFlyout" or "")
	self.wndMain:FindChild("BGmove"):Show(self.settings.bFreePositioning or self.settings.bFreeScaling)
	self.wndMain:SetStyle("Moveable", self.settings.bFreePositioning)
	self.wndMain:SetStyle("Sizable", self.settings.bFreeScaling)
	self.wndMain:SetStyle("IgnoreMouse", not (self.settings.bFreePositioning or self.settings.bFreeScaling))
end

function LocalQuestTrackerAdvanced:OnFreeScalingButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bFreeScaling = wndControl:IsChecked()
	self:ApplyFreePositionSize()
end

function LocalQuestTrackerAdvanced:OnWidthSliderBarChanged( wndHandler, wndControl, fNewValue, fOldValue )
	self.settings.nWidth = fNewValue
	wndControl:GetParent():FindChild("Value"):SetText(fNewValue)
	self:SetSize()
end

function LocalQuestTrackerAdvanced:OnHeightSliderBarChanged( wndHandler, wndControl, fNewValue, fOldValue )
	self.settings.nHeight = fNewValue
	wndControl:GetParent():FindChild("Value"):SetText(fNewValue)
	self:SetSize()
end

function LocalQuestTrackerAdvanced:OnFreePositioningButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bFreePositioning = wndControl:IsChecked()
	self:ApplyFreePositionSize()
end

function LocalQuestTrackerAdvanced:OnPosOffsetXSliderBarChanged( wndHandler, wndControl, fNewValue, fOldValue )
	self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][1] = fNewValue
	wndControl:GetParent():FindChild("Value"):SetText(fNewValue)
	self:SetPosition()
end

function LocalQuestTrackerAdvanced:OnPosOffsetYSliderBarChanged( wndHandler, wndControl, fNewValue, fOldValue )
	self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor][2] = fNewValue
	wndControl:GetParent():FindChild("Value"):SetText(fNewValue)
	self:SetPosition()
end

function LocalQuestTrackerAdvanced:OnTopAlignmentButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bTopAlign = wndControl:IsChecked()
	self:RequestRedrawAll()
end

function LocalQuestTrackerAdvanced:OnMainAnchorGridSelChange(wndHandler, wndControl, nRow)
	local tData = wndControl:GetCellData(nRow, 1)
	if not tData then return end
	self.wndSettingsSizePosition:FindChild("AnchorDropdownParent"):Show(false)
	self.wndSettingsSizePosition:FindChild("Position:Anchor:DropdownButton"):SetText(tData["sLabel"])
	self.wndSettingsSizePosition:FindChild("Position:PosOffsetX:Slider"):SetValue(self.settings.tWindowPositions.tMainOffsets[nRow][1])
	self.wndSettingsSizePosition:FindChild("Position:PosOffsetX:Value"):SetText(self.settings.tWindowPositions.tMainOffsets[nRow][1])
	self.wndSettingsSizePosition:FindChild("Position:PosOffsetY:Slider"):SetValue(self.settings.tWindowPositions.tMainOffsets[nRow][2])
	self.wndSettingsSizePosition:FindChild("Position:PosOffsetY:Value"):SetText(self.settings.tWindowPositions.tMainOffsets[nRow][2])
	self.settings.tWindowPositions.nMainAnchor = nRow
	self:SetPosition()
end

function LocalQuestTrackerAdvanced:OnMainAnchorButtonCheck( wndHandler, wndControl, eMouseButton )
	self.wndSettingsSizePosition:FindChild("AnchorDropdownParent"):Show(not self.bMainAnchorShown, true)
end

function LocalQuestTrackerAdvanced:OnMainAnchorShow( wndHandler, wndControl )
	if wndHandler ~= wndControl then return end
	self:FillAnchorList()
	self.bMainAnchorShown = true
	self.wndSettingsSizePosition:FindChild("Position:Anchor:DropdownButtonBackground"):SetCheck(true)
	self.wndSettingsSizePosition:FindChild("Position:Anchor:DropdownButton:Arrow"):SetSprite("CRB_Basekit:kitIcon_Holo_UpArrow")
	self.wndSettings:SetStyle("Escapable", false)
end

function LocalQuestTrackerAdvanced:OnMainAnchorClosed( wndHandler, wndControl )
	if wndHandler ~= wndControl then return end
	self.bMainAnchorShown = false
	self.wndSettingsSizePosition:FindChild("Position:Anchor:DropdownButtonBackground"):SetCheck(false)
	self.wndSettingsSizePosition:FindChild("Position:Anchor:DropdownButton:Arrow"):SetSprite("CRB_Basekit:kitIcon_Holo_DownArrow")
	Apollo.CreateTimer("MainEscapableTimer", 0.1, false)
end

function LocalQuestTrackerAdvanced:OnMainAnchorButtonEnter( wndHandler, wndControl, x, y )
	self.wndSettingsSizePosition:FindChild("AnchorDropdownParent"):SetStyle("CloseOnExternalClick", false)
end

function LocalQuestTrackerAdvanced:OnMainAnchorButtonExit( wndHandler, wndControl, x, y )
	self.wndSettingsSizePosition:FindChild("AnchorDropdownParent"):SetStyle("CloseOnExternalClick", true)
end

function LocalQuestTrackerAdvanced:OnMainEscapableTimer()
	self.wndSettings:SetStyle("Escapable", true)
end

---------------------------------------------------------------------------------------------------
-- SizePosition Functions
---------------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:SetPosition()
	local nWidth = self.settings.nWidth
	local nHeight = self.settings.nHeight
	local fScale = 1 - (1 / Apollo.GetConsoleVariable("ui.Scale"))
	local nOffsetX, nOffsetY = VILib:Round(math.abs(nWidth * fScale) / 2), VILib:Round(math.abs(nHeight * fScale) / 2)
	self.wndMain:SetAnchorPoints(unpack(tAnchors[self.settings.tWindowPositions.nMainAnchor].tPoints))
	local nX, nY = unpack(self.settings.tWindowPositions.tMainOffsets[self.settings.tWindowPositions.nMainAnchor])
	if tonumber(nX) and tonumber(nY) then
		local nOL, nOT, nOR, nOB = unpack(tAnchors[self.settings.tWindowPositions.nMainAnchor].tOffsetMults)
		local nL, nT, nR, nB = VILib:Round(nWidth * nOL) + nX, VILib:Round(nHeight * nOT) + nY, VILib:Round(nWidth * nOR) + nX, VILib:Round(nHeight * nOB) + nY	
		local nShiftDatachron  = 0
		-- Reposition for Datachron
		if g_wndDatachron and g_wndDatachron:IsShown() and self.settings.tWindowPositions.nMainAnchor == 7 then
			local nL2, nT2, nR2, nB2 = g_wndDatachron:GetAnchorOffsets()
			nShiftDatachron = (nB2 - g_wndDatachron:GetHeight() + 50) - (nB - nOffsetY)
			if Apollo.GetAddon("SMDC") then nShiftDatachron = nShiftDatachron - 50 end
		end
		-- Reposition for ChallengeTracker
		if self.bChallengeVisible and self.settings.tWindowPositions.nMainAnchor == 7 then
		end
		-- Set Position
		self.wndMain:SetAnchorOffsets(nL - nOffsetX, nT - nOffsetY, nR - nOffsetX, nB - nOffsetY + nShiftDatachron)
	end

	self:RequestRedrawAll()
end

function LocalQuestTrackerAdvanced:SetSize()
	self:SetPosition()
end

---------------------------------------------------------------------------------------------------
-- SettingsFilterPreferences Functions
---------------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:ApplyFilterPreferences()
	self:ResetHide(self.tMinimized.tQuestTime, self.tMinimized.tQuest)
	self:ResetHide(self.tMinimized.tEpisodeTime, self.tMinimized.tEpisode)
	log:debug(self.tMinimized.tQuest)
	log:debug(self.tMinimized.tEpisode)
end

function LocalQuestTrackerAdvanced:OnEpisodesCollapseButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bAutoCollapseEpisodes = wndControl:IsChecked()
	self:ApplyFilterPreferences()
end

function LocalQuestTrackerAdvanced:OnQuestsCollapseButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bAutoCollapseQuests = wndControl:IsChecked()
	self:ApplyFilterPreferences()
end

function LocalQuestTrackerAdvanced:OnTimeoutSliderBarChanged( wndHandler, wndControl, fNewValue, fOldValue )
	self.settings.nTimeout = fNewValue
	wndControl:GetParent():FindChild("Value"):SetText(fNewValue)
	self:ApplyFilterPreferences()
end

function LocalQuestTrackerAdvanced:OnWorldStoryTrackButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bAutoTrackWorldStory = wndControl:IsChecked()
	self:AutoTrack()
end

function LocalQuestTrackerAdvanced:OnZoneStoryTrackButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bAutoTrackZoneStory = wndControl:IsChecked()
	self:AutoTrack()
end

function LocalQuestTrackerAdvanced:OnRegionalStoryTrackButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bAutoTrackRegionalStory = wndControl:IsChecked()
	self:AutoTrack()
end

function LocalQuestTrackerAdvanced:OnTasksTrackButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bAutoTrackTasks = wndControl:IsChecked()
	self:AutoTrack()
end

function LocalQuestTrackerAdvanced:OnTradeskillsShowHideButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bAutoShowHideTradeskills = wndControl:IsChecked()
	self:OnCheckCraftingStationDistanceTimer()
end

function LocalQuestTrackerAdvanced:OnEpisodesShowHideButtonUp( wndHandler, wndControl, eMouseButton )
	self.settings.bAutoHideEpisodes = wndControl:IsChecked()
	self:DestroyAndRedraw()
end

---------------------------------------------------------------------------------------------------
-- FilterPreferences Functions
---------------------------------------------------------------------------------------------------

function LocalQuestTrackerAdvanced:GetDistance(target)
	if GameLib.GetPlayerUnit() ~= nil then
		local playerPos = GameLib.GetPlayerUnit():GetPosition()
		local playerVec = Vector3.New(playerPos.x, playerPos.y, playerPos.z)
		if Vector3.Is(target) then
			return (playerVec - target):Length()
		elseif Unit.is(target) then
			local targetPos = target:GetPosition()
			if targetPos == nil then
				return 0 -- Uh, no clue
			end
			local targetVec = Vector3.New(targetPos.x, targetPos.y, targetPos.z)
			return (playerVec - targetVec):Length()
		else
			local targetVec = Vector3.New(target.x, target.y, target.z)
			return (playerVec - targetVec):Length()
		end
	else
		return nil
	end
end

function LocalQuestTrackerAdvanced:OnCheckCraftingStationDistanceTimer()
	local closestCraftingStation = nil
	local closestCraftingStationDist = 0xffffff
	local dist = nil
	for _, cS in pairs(self.craftingStationList) do
		dist = self:GetDistance(cS)
		if dist < closestCraftingStationDist then
			closestCraftingStation = cS
			closestCraftingStationDist = dist
		end
	end

	
	if (closestCraftingStationDist <= 5 and not self.showOnlyTradeskills) or (closestCraftingStationDist > 5 and self.showOnlyTradeskills) then
		self.showOnlyTradeskills = not self.showOnlyTradeskills
		self:DestroyAndRedraw()
		log:debug(self.showOnlyTradeskills)
	end
end

function LocalQuestTrackerAdvanced:OnUnitCreated(unit)
	if unit:GetType() == 'Simple' and (
										unit:GetName() == 'Crafting Station' or
										unit:GetName() == 'Cooking Grill'
	) then
		self.craftingStationList[unit:GetId()] = unit
	end
end

function LocalQuestTrackerAdvanced:OnUnitDestroyed(unit)
	if unit:GetType() == 'Simple' and (
										unit:GetName() == 'Crafting Station' or
										unit:GetName() == 'Cooking Grill'
	) then
		self.craftingStationList[unit:GetId()] = nil
	end
end

function LocalQuestTrackerAdvanced:DoesCategoryEpisodeHaveQuests(qcCategory, epiEpisode, tCategoryEpisodeHaveQuestsCache)
	local strEpisodeKey = "C"..qcCategory:GetId().."E"..epiEpisode:GetId()
	if  tCategoryEpisodeHaveQuestsCache[strEpisodeKey] ~= nil then
		return tCategoryEpisodeHaveQuestsCache[strEpisodeKey]
	end

	if not epiEpisode:IsWorldStory() then
		for idx, queQuest in pairs(epiEpisode:GetAllQuests(qcCategory:GetId())) do
			if (queQuest:GetState() ~= Quest.QuestState_Completed) and not queQuest:IsIgnored() then
				tCategoryEpisodeHaveQuestsCache[strEpisodeKey] = true
				return true
			end
		end
	end
	tCategoryEpisodeHaveQuestsCache[strEpisodeKey] = false
	return false
end

function LocalQuestTrackerAdvanced:DoesCategoryHaveQuests(qcCategory, tCategoryHaveQuestsCache)
	local tCategoryEpisodeHaveQuestsCache = {}
	local strCategoryKey = "C"..qcCategory:GetId()
	if tCategoryHaveQuestsCache[strCategoryKey] ~= nil then
		return tCategoryHaveQuestsCache[strCategoryKey]
	end

	for idx, epiEpisode in pairs(qcCategory:GetEpisodes()) do
		if self:DoesCategoryEpisodeHaveQuests(qcCategory, epiEpisode, tCategoryEpisodeHaveQuestsCache) then
			tCategoryHaveQuestsCache[strCategoryKey] = true
			return true
		end
	end
	tCategoryHaveQuestsCache[strCategoryKey] = false
	return false
end

function LocalQuestTrackerAdvanced:AutoTrack()
	if self.settings.bAutoTrackWorldStory or self.settings.bAutoTrackZoneStory or self.settings.bAutoTrackRegionalStory or self.settings.bAutoTrackTasks then
		for idx, epiEpisode in pairs(QuestLib.GetTrackedEpisodes(self.bQuestTrackerByDistance)) do
			-- Clear World Story Quests
			if self.settings.bAutoTrackWorldStory and epiEpisode:IsWorldStory() then
				for key, queQuest in pairs(epiEpisode:GetTrackedQuests()) do
					self:QueueQuestForDestroy(queQuest)
					queQuest:SetTracked(false)
				end
			-- Clear Zone Story Quests
			elseif self.settings.bAutoTrackZoneStory and epiEpisode:IsZoneStory() then
				for key, queQuest in pairs(epiEpisode:GetTrackedQuests()) do
					self:QueueQuestForDestroy(queQuest)
					queQuest:SetTracked(false)
				end
			-- Clear Zone Story Quests
			elseif self.settings.bAutoTrackRegionalStory and epiEpisode:IsRegionalStory() then
				for key, queQuest in pairs(epiEpisode:GetTrackedQuests()) do
					self:QueueQuestForDestroy(queQuest)
					queQuest:SetTracked(false)
				end
			-- Clear Task Quests
			elseif self.settings.bAutoTrackTasks and not epiEpisode:IsWorldStory() and not epiEpisode:IsZoneStory() and not epiEpisode:IsRegionalStory() then
				for key, queQuest in pairs(epiEpisode:GetTrackedQuests()) do
					self:QueueQuestForDestroy(queQuest)
					queQuest:SetTracked(false)
				end
			end
		end
		if self.currentZone == nil then
			self.currentZone = GameLib.GetCurrentZoneMap().strName
		end
		local tCategoryEpisodeHaveQuestsCache = {}
		local tCategoryHaveQuestsCache = {}
		for idx1, qcCategory in pairs(QuestLib.GetKnownCategories()) do
			if self:DoesCategoryHaveQuests(qcCategory, tCategoryHaveQuestsCache) then
				for idx2, epiEpisode in pairs(qcCategory:GetEpisodes()) do
					if self:DoesCategoryEpisodeHaveQuests(qcCategory, epiEpisode, tCategoryEpisodeHaveQuestsCache) and (epiEpisode:GetZoneName() == "" or epiEpisode:GetZoneName() == self.currentZone) then
						-- Track World Story Quests
						if self.settings.bAutoTrackWorldStory and epiEpisode:IsWorldStory() then
							for idx3, queQuest in pairs(epiEpisode:GetAllQuests(qcCategory:GetId())) do
								if queQuest:GetState() ~= Quest.QuestState_Completed and not queQuest:IsIgnored() then
									queQuest:SetTracked(true)
								end
							end
						-- Track Zone Story Quests
						elseif self.settings.bAutoTrackZoneStory and epiEpisode:IsZoneStory() then
							for idx3, queQuest in pairs(epiEpisode:GetAllQuests(qcCategory:GetId())) do
								if queQuest:GetState() ~= Quest.QuestState_Completed and not queQuest:IsIgnored() then
									queQuest:SetTracked(true)
								end
							end
						-- Track Zone Story Quests
						elseif self.settings.bAutoTrackRegionalStory and epiEpisode:IsRegionalStory() then
							for idx3, queQuest in pairs(epiEpisode:GetAllQuests(qcCategory:GetId())) do
								if queQuest:GetState() ~= Quest.QuestState_Completed and not queQuest:IsIgnored() then
									queQuest:SetTracked(true)
								end
							end
						-- Track Task Quests
						elseif self.settings.bAutoTrackTasks and not epiEpisode:IsWorldStory() and not epiEpisode:IsZoneStory() and not epiEpisode:IsRegionalStory() then
							for idx3, queQuest in pairs(epiEpisode:GetAllQuests(qcCategory:GetId())) do
								if queQuest:GetState() ~= Quest.QuestState_Completed and not queQuest:IsIgnored() and queQuest:GetCategory() ~= nil and queQuest:GetCategory():GetTitle() == self.currentZone then
									queQuest:SetTracked(true)
								end
							end
						end
					end
				end
			end
		end
		self:DestroyAndRedraw()
	end
end


---------------------------------------------------------------------------------------------------
-- Initialize Local Quest Tracker Advanced Instance
---------------------------------------------------------------------------------------------------

local LocalQuestTrackerAdvancedInst = LocalQuestTrackerAdvanced:new()
LocalQuestTrackerAdvancedInst:Init()