require "ChatSystemLib"


local VILib = {}


function VILib:Round(n, nDecimals, bForceDecimals)
	local nMult = 1
	if nDecimals and nDecimals > 0 then
		for i = 1, nDecimals do nMult = nMult * 10 end
	end
	n = math.floor((n * nMult) + 0.5) / nMult
	if bForceDecimals then n = string.format("%." .. nDecimals .. "f", n) end
	return n
end

function VILib:ConvertColorToHex(color)
	return string.format("%02x%02x%02x%02x", self:Round(255 * color.a), self:Round(255 * color.r), self:Round(255 * color.g), self:Round(255 * color.b))
end

function VILib:ConvertHexToColor(hex)
	local a = string.sub(hex, 1, 2)
	hex = string.sub(hex, -6)
	local r, g, b = 0, 0, 0
	local n = tonumber(hex, 16)
	if n then
		r = math.floor(n / 65536)
		g = math.floor((n / 256) % 256)
		b = n % 256
	end
	return r / 255, g / 255, b / 255, 1
end

function VILib:CopyTable(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == "table" then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[self:CopyTable(orig_key)] = self:CopyTable(orig_value)
        end
        setmetatable(copy, self:CopyTable(getmetatable(orig)))
    else
        copy = orig
    end
    return copy
end

function VILib:MergeTables(t1, t2)
    for k, v in pairs(t2) do
    	if type(v) == "table" then
			if t1[k] then
	    		if type(t1[k] or false) == "table" then
	    			self:MergeTables(t1[k] or {}, t2[k] or {})
	    		else
	    			t1[k] = v
	    		end
			else
				t1[k] = {}
    			self:MergeTables(t1[k] or {}, t2[k] or {})
			end
    	else
    		t1[k] = v
    	end
    end
    return t1
end

function VILib:WriteToChat(text)
	ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Command, text)
end


_G["GeminiPackages"]:NewPackage(VILib, "VILib-1.2", 1)
